import asyncio
import dataclasses
import enum
import io
import logging
import os
import signal
import traceback
import uuid
from collections import deque
from datetime import datetime
from typing import Dict, List, Optional, Union

import aiomultiprocess
import setproctitle
import zmq.error

from .experiment_run import ExperimentRun
from .run_governor import RunGovernor
from ..core import MajorDomoBroker, MajorDomoClient, RuntimeConfig
from ..core.protocol import (
    ExperimentRunShutdownRequest,
    ExperimentRunShutdownResponse,
    ExperimentRunStartRequest,
    ExperimentRunStartResponse,
)
from ..util import LogServer
from ..util import spawn_wrapper

LOG = logging.getLogger(__name__)


class ExecutorState(enum.Enum):
    PRISTINE = 0
    INITIALIZED = 1
    RUNNING = 2
    SHUTDOWN = 3
    EXITED = 4
    SIGINT = 5
    SIGABRT = 6
    SIGTERM = 7


@dataclasses.dataclass
class _RunGovernorPCB:
    """Simple Process Control Block for RunGovernor control."""

    run_governor_id: str
    started_at: datetime
    run_governor_process: aiomultiprocess.Process
    experiment_run: ExperimentRun
    experiment_run_id: Optional[str] = None


@dataclasses.dataclass
class ExperimentRunRuntimeInformation:
    """Accumulated information about the one experiment run

    This structure contains information about one experiment run.
    It stores on which :py:class:`RunGovernor` it is executed, when it
    was started, which run is being executed and what ID the
    experiment run has.
    """

    experiment_run: ExperimentRun
    started_at: Union[datetime, None] = None
    run_governor_id: Union[str, None] = None
    experiment_run_id: Union[str, None] = None

    @property
    def is_running(self):
        return self.experiment_run_id is not None


class ExperimentRunStartError(RuntimeError):
    def __init__(self, experiment_run_id, run_governor_id, message):
        super().__init__(message)
        self.experiment_run_id = experiment_run_id
        self.run_governor_id = run_governor_id
        self.message = message


class DeadMajorDomoBrokerError(RuntimeError):
    def __init__(self):
        super().__init__()


class InterruptSignal(RuntimeError):
    def __init__(self):
        super().__init__()


async def _execute_run_governor(uid: str, broker_uri: str):
    """Executes the ::`RunGovernor` main loop, catching errors

    This is a wrapper function around that creates a ::`RunGovernor` and
    handles ::`RunGovernor.run`. It takes care of clearing signal handlers,
    setting the proctitle, and generally catching errors in a meaningful in
    order to report it to the Executor without simply dying.

    This method belongs to the ::`Executor` logically, but is not part of the
    class in order to avoid serialization/deserialization of the whole Executor
    each time a new ::`RunGovernor` process is spawned.

    Parameters
    ----------
    uid : str
        UID of the new ::`RunGovernor`
    broker_uri : str
        URI of the ::`MajorDomoBroker` instance to connect to for further
        communication

    Return
    ------
    Nothing.
    """
    setproctitle.setproctitle("palaestrAI[RunGovernor-%s]" % uid[-6:])
    signal.signal(signal.SIGINT, signal.default_int_handler)
    signal.signal(signal.SIGABRT, signal.SIG_DFL)
    signal.signal(signal.SIGTERM, signal.SIG_DFL)
    try:
        run_gov = RunGovernor(
            uid=uid,
            broker_uri=broker_uri,
        )
        await run_gov.run()
    except Exception as e:
        LOG.critical("Execution of RunGovernor(uid=%s) failed: %s.", uid, e)
        tb = io.StringIO()
        traceback.print_exc(file=tb)
        LOG.debug(
            "Execution of RunGovernor(uid=%s) failed: %s; stack trace "
            "is: %s.",
            uid,
            e,
            tb.getvalue(),
        )
        raise


class Executor:
    """The executor is the entrypoint for every run execution.

    The role of the executor is to receive new experiment runs and
    distribute them to existing :class:`RunGovernor` instances. If
    palaestrAI is used in local run mode, the executor will initialize
    a :class:`RunGovernor`.

    Furthermore, the Executor can stop running experiment runs.

    Notes
    -----
        At some point, when the core protocol has progressed far enough,
        we will be able to run several experiments at once from an
        executor. But, until we're sure we can, the public API accepts
        only one experiment.

    """

    def __init__(self):
        # (At some point, when the core protocol has progressed far enough,
        # we will be able to run several experiments at once from an executor.
        # But, until we're sure we can, the public API accepts only one
        # experiment.)

        self._state: ExecutorState = ExecutorState.PRISTINE
        self._broker: Union[None, MajorDomoBroker] = None
        self._client: Union[None, MajorDomoClient] = None
        self._broker_process: Union[None, aiomultiprocess.Process] = None
        self._broker_uri = "tcp://%s:%s" % (
            "*" if RuntimeConfig().public_bind else "127.0.0.1",
            RuntimeConfig().executor_bus_port,
        )
        self._run_governors: Dict[str, _RunGovernorPCB] = {}
        self._runs_scheduled = deque()
        self._log_server = LogServer("127.0.0.1", RuntimeConfig().logger_port)
        aiomultiprocess.set_start_method(
            "spawn"
        )  # Set best default for this OS.

    def _handle_signal_interrupt(self, signum):
        """Handle interrupting signals by notifying of the state change."""
        LOG.info("Executor has received signal %s, " "shutting down", signum)
        if signum not in {signal.SIGABRT, signal.SIGINT, signal.SIGTERM}:
            return
        old_state = self._state
        self._state = {
            signal.SIGINT.value: ExecutorState.SIGINT,
            signal.SIGABRT.value: ExecutorState.SIGABRT,
            signal.SIGTERM.value: ExecutorState.SIGTERM,
        }[signum]
        LOG.info(
            "Executor changed state from %s to %s.", old_state, self._state
        )

    def _init_signal_handler(self):
        """Sets handlers for interrupting signals in the event loop."""
        signals = {signal.SIGINT, signal.SIGABRT, signal.SIGTERM}
        LOG.debug(
            "Executor(id=0x%x) registering signal handlers for signals %s.",
            id(self),
            signals,
        )
        loop = asyncio.get_running_loop()
        for signum in signals:
            loop.add_signal_handler(
                signum, self._handle_signal_interrupt, signum
            )

    async def _init_logging(self):
        """Starts the log server and configures log filters"""
        await self._log_server.start()

        def filter_record_above_debug(record: logging.LogRecord):
            if record.levelname != "DEBUG":
                return False
            return True

        for h in [h for h in logging.root.handlers if "debug" in h.name]:
            h.addFilter(filter_record_above_debug)

    def _init_communication(self):
        """Initialization of all core components"""
        self._broker = MajorDomoBroker(self._broker_uri)
        self._broker_process = aiomultiprocess.Process(
            daemon=True,
            target=spawn_wrapper,
            args=(
                "MajorDomoBroker",
                RuntimeConfig().to_dict(),
                self._broker.mediate,
            ),
            name="Executor-MajorDomoBroker",
        )
        self._broker_process.start()
        LOG.debug(
            "Executor(id=0x%x) started MajorDomoBroker(id=0x%x, uri=%s).",
            id(self),
            id(self._broker),
            self._broker_uri,
        )
        self._client = MajorDomoClient(
            "tcp://127.0.0.1:%s" % RuntimeConfig().executor_bus_port
        )
        LOG.debug(
            "Executor(id=0x%x) started MajorDomoClient(id=0x%x, uri=%s).",
            id(self),
            id(self._client),
            "tcp://127.0.0.1:%s" % RuntimeConfig().executor_bus_port,
        )

    @staticmethod
    def _join_process(process: aiomultiprocess.Process):
        while process.is_alive():
            try:
                process.join(3)
            except TimeoutError:
                pass

    def experiment_runs(self) -> List[ExperimentRunRuntimeInformation]:
        runs = list()
        for run in self._runs_scheduled:
            runs.append(ExperimentRunRuntimeInformation(experiment_run=run))

        for pcb in self._run_governors.values():
            runs.append(
                ExperimentRunRuntimeInformation(
                    experiment_run=pcb.experiment_run,
                    started_at=pcb.started_at,
                    experiment_run_id=pcb.experiment_run_id,
                )
            )
        return runs

    async def _launch_run_governor(
        self, experiment_run: ExperimentRun
    ) -> _RunGovernorPCB:
        """Launches a new :py:class:`RunGovernor` process

        This method creates a new `aiomultiprocess.Process` and launches it,
        returning the appropriate PCB object.

        :rtype: _RunGovernorPCB
        """
        run_gov_uid = str(uuid.uuid4())
        run_gov_process = aiomultiprocess.Process(
            target=spawn_wrapper,
            args=(
                "RunGovernor-%s" % run_gov_uid,
                RuntimeConfig().to_dict(),
                _execute_run_governor,
                [run_gov_uid, self._broker_uri],
            ),
        )
        run_gov_process.start()

        return _RunGovernorPCB(
            experiment_run=experiment_run,
            started_at=datetime.now(),
            run_governor_id=run_gov_uid,
            run_governor_process=run_gov_process,
        )

    async def _deploy_experiment_run(self, experiment_run: ExperimentRun):
        """Creates a :py:class:`RunGovernor` for a given experiment
        run.

        This method creates a new :py:class:`RunGovernor` and deploys
        the provided :py:class:`ExperimentRun` to it, pending start.

        The new :py:class:`RunGovernor` is started as a "python-
        parallel" process. All information about the running process is
        saved in the `self._run_governors` list, where we can also take
        care to stop all run governors should we need to.

        Parameters
        ----------
        experiment_run: ExperimentRun
            The experiment that should be executed.

        Returns
        -------
        str
            An opaque :py:class:`RunGovernor` identification string.
        """

        pcb = await self._launch_run_governor(experiment_run)
        self._run_governors[pcb.run_governor_id] = pcb
        LOG.debug(
            "Executor(id=0x%x) started RunGovernor(uid=%s, broker_uri=%s)",
            id(self),
            pcb.run_governor_id,
            self._broker_uri,
        )
        return pcb.run_governor_id

    async def _start_experiment_run(self, run_governor_id):
        """Starts a previously deployed experiment run.

        When the experiment run is deployed via
        :py:meth:`_deploy_experiment_run`, the :py:class:`RunGovernor`
        is ready, but does not yet start it. This is done through this
        method, which sends the appropriate messages over the executor
        communications bus.

        Parameters
        ----------
        run_governor_id: str
            ID of the :class:`RunGovernor` that shall commence the
            experiment run.
        """
        experiment_run_id = str(uuid.uuid4())
        pcb = self._run_governors[run_governor_id]
        msg = ExperimentRunStartRequest(
            sender_executor_id=id(self),
            receiver_run_governor_id=run_governor_id,
            experiment_run_id=experiment_run_id,
            experiment_run=pcb.experiment_run,
        )
        LOG.debug(
            "Executor(id=%s) starting experiment run: "
            "ExperimentStartRequest(user_id='%s', run_id=%s) "
            " - Our business is life itself",
            id(self),
            pcb.experiment_run.user_id,
            experiment_run_id,
        )
        response = await self._client.send(run_governor_id, msg)

        if isinstance(response, ExperimentRunStartResponse):
            if response.successful is True:
                pcb.experiment_run_id = experiment_run_id
                LOG.debug(
                    "Executor(id=0x%x) received ExperimentRunStart"
                    "Response from RunGovernor(uid=%s) for "
                    "ExperimentRun(run_id=%s).",
                    id(self),
                    run_governor_id,
                    experiment_run_id,
                )
            else:
                LOG.error("Could not start experiment run: %s", response.error)
                raise ExperimentRunStartError(
                    experiment_run_id=experiment_run_id,
                    run_governor_id=run_governor_id,
                    message=response.error,
                )
        else:
            LOG.error(
                "Executor expected ExperimentRunStartResponse, but got "
                "'%s' instead",
                response,
            )
            raise TypeError
        LOG.debug(
            "Executor(id=0x%x) started experiment run (experiment_run_id=%s), "
            "got ExperimentStartResponse(successful=%s).",
            id(self),
            experiment_run_id,
            response.successful,
        )
        return experiment_run_id

    async def cancel(self, experiment_run_id):
        """Shuts an experiment run down prematurely.

        This method sends a :py:class:`ExperimentShutdownRequest` to
        the :py:class:`RunGovernor` responsible for executing the
        associated experiment run. This allows for a graceful, yet
        premature shutdown of a running experiment run. Normally,
        experiment runs terminate when their termination condition is
        met, so this method provides a way for an external user to
        interfere with a running experiment.

        Parameters
        ----------
        run_id: str
            UID of the experiment run to shut down.
        """
        rg_dict_item = next(
            filter(
                lambda i: i[1].experiment_run_id == experiment_run_id,
                self._run_governors.items(),
            ),
            None,
        )
        run_governor_id = rg_dict_item[0]

        if run_governor_id is None:
            LOG.error(
                "Executor (id=0x%x) cannot terminate ExperimentRun("
                "run_id=%s): Cannot find RunGovernor",
                id(self),
                experiment_run_id,
            )
            return

        LOG.debug(
            "Executor (id=0x%x) sending ExperimentRunShutdownRequest for "
            "ExperimentRun(run_id=%s).",
            id(self),
            experiment_run_id,
        )
        msg = ExperimentRunShutdownRequest(
            sender_executor_id=id(self),
            receiver_run_governor_id=run_governor_id,
            experiment_run_id=experiment_run_id,
        )
        response = await self._client.send(run_governor_id, msg)
        LOG.debug("Executor(id=0x%x) received %s", id(self), response)
        if not isinstance(response, ExperimentRunShutdownResponse):
            return
        if not response.successful:
            # This may lead to an error when performed during an
            # iteration over the run_governors:
            # self._run_governors.pop(run_governor_uid, None)
            raise RuntimeError(response.error)

    def schedule(self, experiment_run: ExperimentRun):
        """Schedules an experiment run to be executed."""
        self._runs_scheduled.append(experiment_run)

    def shutdown(self):
        """Performs an oderly shutdown"""
        # We need to discriminate via state. If we're currently running, but
        # receive a call to "shutdown", then its coming from outside and we
        # need to change the flag to have an orderly shutdown on the main loop
        # in .run().
        # Otherwise, we do the actual shutdown and cleanup.

        if self._state.value > ExecutorState.RUNNING.value:
            LOG.warning(
                "Executor(id=0x%x) is already shuttdown down.", id(self)
            )
            return  # We're already shutting down.
        LOG.info(
            "Executor(id=0x%x) received command to shut down, changing "
            "state.",
            id(self),
        )
        self._runs_scheduled.clear()
        self._state = ExecutorState.SHUTDOWN

    async def _monitor_state(self):
        known_state = self._state
        while known_state.value == self._state.value:
            try:
                await asyncio.sleep(0.5)
            except asyncio.CancelledError:
                break
        LOG.debug(
            "Executor(id=0x%x) state changed from %s to %s",
            id(self),
            known_state,
            self._state,
        )

    async def _check_major_domo_broker_state(self):
        if (
            self._state != ExecutorState.RUNNING
            or self._broker_process.is_alive()
        ):
            return
        n_broker_checks = 0
        if n_broker_checks == 3:
            LOG.fatal(
                "Executor(id=%s) lost the MajorDomoBroker, "
                "committing seppuku.",
                id(self),
            )
            raise DeadMajorDomoBrokerError()
        else:
            wait_time = 1 + n_broker_checks
            LOG.debug(
                "MajorDomoBroker is not yet online, waiting for" "%s seconds",
                wait_time,
            )
            n_broker_checks += 1
            await asyncio.sleep(wait_time)

    async def _start_all_scheduled_experiments(self):
        """Starts all experiments that are scheduled, monitoring the start

        All experiments that are currently scheduled via
        ::`Executor.schedule` are started. This method initializes a new
        ::`RunGovernor`, deploy the experiment to it, signals the
        ::`RunGovernor` to start the experiment and waits for positive
        response to this ::`ExperimentStartRequest`. If no such response
        arrives, the ::`RunGovernor` is killed and the experiment run
        cancelled.

        The method returns when all scheduled experiments have successfully
        started. It does not keep track of the overall experimentation
        process.
        """
        LOG.debug(
            "Executor(id=0x%x) checks whether to start new experiments.",
            id(self),
        )
        while (
            self._state == ExecutorState.RUNNING
            and len(self._runs_scheduled) > 0
        ):
            experiment = self._runs_scheduled.pop()
            try:
                run_governor_id = await self._deploy_experiment_run(experiment)
                start_experiment_run_task = asyncio.create_task(
                    self._start_experiment_run(run_governor_id)
                )
                process_monitor_task = asyncio.create_task(
                    self._run_governors[
                        run_governor_id
                    ].run_governor_process.join()
                )
                tasks_done, _ = await asyncio.wait(
                    {
                        start_experiment_run_task,
                        process_monitor_task,
                    },
                    return_when=asyncio.FIRST_COMPLETED,
                )
                if (
                    start_experiment_run_task in tasks_done
                    and start_experiment_run_task.exception()
                ):
                    raise start_experiment_run_task.exception()
                if process_monitor_task in tasks_done:
                    raise ExperimentRunStartError(
                        self._run_governors[run_governor_id].experiment_run_id,
                        run_governor_id,
                        "No ExperimentStartResponse received",
                    )
            except (ExperimentRunStartError, asyncio.CancelledError) as e:
                LOG.fatal(
                    "Executor(id=%s) "
                    "could not launch ExperimentRun(experiment_id=%s): "
                    "%s; killing associated RunGovernor(uid=%s).",
                    id(self),
                    experiment.id,
                    e,
                    e.run_governor_id,
                )
                pcb = self._run_governors[e.run_governor_id]
                pcb.run_governor_process.terminate()
                try:
                    await pcb.run_governor_process.join(3)
                except (asyncio.TimeoutError, asyncio.CancelledError):
                    if pcb.run_governor_process.is_alive():
                        pcb.run_governor_process.kill()
                        await pcb.run_governor_process.join(3)

    async def execute(self):
        """Executes an experiment run.

        This method starts the whole experiment run execution process.
        It initializes the :class:`RunGovernor` for the experiment run
        and sets up communication. This method returns only if:

        1. The experiment run has terminated successfully;
        2. an error has occurred, in which case an exception is raised;
        3. the user has terminated the process (e.g., by hitting ^C).

        Returns
        -------
        ExecutorState
            The state the executor is now in, either "SHUTDOWN" if
            everything exited normally, or one of the SIG* states if a
            signal was received.
        """
        setproctitle.setproctitle("palaestrAI[Executor]")
        self._init_communication()
        self._init_signal_handler()
        await self._init_logging()
        self._state = ExecutorState.INITIALIZED

        LOG.debug("Executor(id=0x%x) starting main execution loop", id(self))
        self._state = ExecutorState.RUNNING
        state_change_monitor_task = asyncio.create_task(self._monitor_state())
        while self._state == ExecutorState.RUNNING:
            try:
                await self._check_major_domo_broker_state()
            except DeadMajorDomoBrokerError:
                self._state = ExecutorState.SHUTDOWN
                continue

            start_experiments_task = asyncio.create_task(
                self._start_all_scheduled_experiments()
            )
            tasks_done, tasks_pending = await asyncio.wait(
                {state_change_monitor_task, start_experiments_task},
                return_when=asyncio.FIRST_COMPLETED,
            )
            if (
                state_change_monitor_task in tasks_done
                and start_experiments_task in tasks_pending
            ):
                start_experiments_task.cancel()

            LOG.debug(
                "Executor(id=0x%x) checks for finished RunGovernors",
                id(self),
            )
            self._run_governors = {
                uid: pcb
                for uid, pcb in self._run_governors.items()
                if pcb.run_governor_process.is_alive()
            }
            LOG.info(
                "Executor(id=0x%x) has %d active run governors and %d "
                "experiment runs scheduled.",
                id(self),
                len(self._run_governors),
                len(self._runs_scheduled),
            )
            await asyncio.sleep(3)
            if (
                self._state == ExecutorState.RUNNING
                and len(self._run_governors) + len(self._runs_scheduled) == 0
            ):
                self._state = ExecutorState.SHUTDOWN
        state_change_monitor_task.cancel()
        await self._shutdown()
        return self._state

    async def _shutdown_all_run_governors(self):
        """Shuts all ::`RunGovernor` instances down, cleaning up

        This method terminates all running ::`RunGovernor` instances. They
        get a nice message first, but are forcefully terminated if they don't
        react. Also, the internal data structures of the ::`Executor` are
        cleaned up.
        """
        for run_gov_uid in self._run_governors:
            pcb = self._run_governors[run_gov_uid]
            if not pcb.run_governor_process.is_alive():
                continue
            LOG.debug(
                "Executor(id=%s) "
                "signalling RunGovernor(uid=%s, run_id=%s) "
                "to shut down.",
                id(self),
                run_gov_uid,
                pcb.experiment_run_id,
            )
            if pcb.experiment_run_id:
                try:
                    await asyncio.wait_for(
                        self.cancel(pcb.experiment_run_id), timeout=15
                    )
                    continue
                except asyncio.TimeoutError:
                    LOG.debug(
                        "Executor(id=%s) has encountered a "
                        "RunGovernor(uid=%s, run_id=%s) "
                        "that seems to be still active, trying to abort. "
                        "Did you hit ^C? Oh, you bad boy...",
                        id(self),
                        pcb.run_governor_id,
                        pcb.experiment_run_id,
                    )
            LOG.warning(
                "Executor(id=0x%x) "
                "could not shut down "
                "RunGovernor(uid=%s, run_id=%s) "
                "orderly.",
                id(self),
                run_gov_uid,
                pcb.experiment_run_id,
            )
            run_governor_pgid = os.getpgid(pcb.run_governor_process.pid)
            pcb.run_governor_process.terminate()
            try:
                await pcb.run_governor_process.join(3)
            except asyncio.TimeoutError:
                pcb.run_governor_process.kill()
            try:
                # Zombies will be shot:
                os.killpg(run_governor_pgid, signal.SIGKILL)
            except ProcessLookupError:
                pass  # Just a shot in the dark.

    async def _shutdown(self):
        LOG.info("Executor(id=%s) initiating shutdown procedure.", id(self))
        await self._shutdown_all_run_governors()
        try:
            await self._client.destroy()
        except zmq.error.ZMQError:
            # This can happen on ^C. It's actually not that bad, so we just
            # do a debug log entry here.
            LOG.debug(
                "Executor(id=%s) "
                "could not send destroy message via MajorDomoClient to "
                "MajorDomoBroker(uri=%s); ignoring that anyways and dragging "
                "on.",
                id(self),
                self._broker_uri,
            )
        self._broker_process.terminate()
        try:
            await self._broker_process.join(3)
        except asyncio.TimeoutError:
            self._broker_process.kill()
        if self._state == ExecutorState.SHUTDOWN:
            self._state = ExecutorState.EXITED
        await self._log_server.stop()



"""reward to table muscle_action

Revision ID: 158c30d6a4bf
Revises: f29bac335186
Create Date: 2020-11-26 14:13:42.943147

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "158c30d6a4bf"
down_revision = "f29bac335186"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("muscle_actions", sa.Column("reward", sa.FLOAT))


def downgrade():
    op.drop_column("muscle_actions", "reward")



import os
from typing import TYPE_CHECKING

from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from palaestrai.experiment import RunGovernor


class RGSPristine(RunGovernorState):
    """Represent the PRISTINE state of the run governor.

    No changes on the run governor have taken place until :meth:`run`
    is called. During this state, the process group is set so that
    all subprocesses can be identified as subprocesses of the run
    governor on the OS level.

    Notes
    -----
    The operation os.setpgrp is only available on UNIX-like systems.
    If a port to windows is ever considered again, a workaround is
    required.

    Possible next states are:

    * :class:`.RGSInitializing` since no errors are expected in this
      state.

    Parameters
    ----------
    rgc: :class:`.RunGovernor`
        The run governor instance that provides the context for this
        state.


    """

    def __init__(self, rgc: "RunGovernor"):
        super().__init__(rgc, "PRISTINE")

    async def run(self):
        # TODO: Maybe do some further cleanup first
        self.rgc.experiment_run = None
        self.rgc.experiment_run_id = None
        self.rgc.last_request = list()
        self.rgc.next_reply = list()

        os.setpgrp()
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) now leads process group %s.",
            id(self.rgc),
            self.rgc.uid,
            os.getpgid(os.getpid()),
        )

    def next_state(self):
        from . import RGSInitializing

        self.rgc.state = RGSInitializing(self.rgc)



import time

from sqlalchemy import create_engine
from sqlalchemy.exc import ProgrammingError, OperationalError
from sqlalchemy.sql.expression import text
from sqlalchemy_utils import database_exists, create_database

from . import LOG
from .database_model import Model


def _create_timescaledb_extension(engine):
    with engine.begin() as conn:
        conn.execute("CREATE EXTENSION IF NOT EXISTS timescaledb CASCADE;")
    timescale_tables = {
        "world_states",
        "muscle_actions",
        "muscle_sensor_readings",
        "muscle_states",
    }
    with engine.begin() as conn:
        for tbl in timescale_tables:
            cmd = (
                "SELECT * FROM create_hypertable("
                "'%s', 'id', chunk_time_interval => 40960);" % tbl
            )
            res = conn.execute(text(cmd))
            LOG.debug(
                'Result of executing "%s" during setup: %s',
                cmd,
                res.fetchall(),
            )
            res.close()


def setup_database(uri):
    """Creates the database from the current model in one go.

    :param uri: The complete database connection URI.
    """
    engine = create_engine(uri)
    while not database_exists(uri):
        i = 1
        if i > 3:  # Hardcoded max tries. No real reason to configure this.
            LOG.critical(
                "Could not create the database. See errors above for more "
                "details. Giving up now."
            )
            raise RuntimeError("Could not create database")
        try:
            create_database(uri)
        except OperationalError as e:
            try:
                import psycopg2.errors

                if isinstance(e.orig, psycopg2.errors.ObjectInUse):
                    LOG.warning(
                        "Could not create database because the template was "
                        "in use. Retrying in %d seconds.",
                        i,
                    )
                    time.sleep(i)
                else:
                    break
            except ImportError:
                pass
        except ProgrammingError as e:
            LOG.error(
                "There was an error creating the database. I will continue "
                "and hope for the best. The error was: %s",
                e,
            )
        i += 1

    with engine.begin() as conn:
        try:
            Model.metadata.create_all(engine)
        except ProgrammingError as e:
            LOG.error("Could not create database: %s" % e)
            raise e
    try:
        _create_timescaledb_extension(engine)
    except OperationalError as e:
        LOG.warning(
            "Could not create extension timescaledb and create hypertables: "
            "%s. "
            "Your database setup might lead to noticeable slowdowns with "
            "larger experiment runs. Please upgrade to PostgreSQL with "
            "TimescaleDB for the best performance." % e
        )


def get_connection(uri):
    """This function creates a database if necessary and returns an engine
    object from SQLAlchemy. The database is automatically created, if not
    existent.

    Args:
        uri (String): <driver>://<user>:<password>@<uri>/<database name>
                e.g.: postgres://otto:vBismarck@8.8.8.8/rent

    Returns:
        engine: A SQLAlchemy object that provides a connection to the database,
        given in the url.
    """
    setup_database(uri)
    return create_engine(uri)



"""This module contains the :class:`.SimulationController`, which is
the abstract base class (ABC) for all specific simulation controllers.
"""
import asyncio
import signal
import uuid
from abc import ABC, abstractmethod
from importlib import import_module

import setproctitle

from palaestrai.agent.agent import Agent
from palaestrai.core import MajorDomoClient, MajorDomoWorker
from palaestrai.core.protocol import (
    AgentSetupRequest,
    AgentShutdownRequest,
    AgentShutdownResponse,
    EnvironmentSetupRequest,
    EnvironmentShutdownRequest,
    EnvironmentShutdownResponse,
    EnvironmentStartRequest,
    EnvironmentStartResponse,
    ShutdownRequest,
    SimulationStartRequest,
    SimulationStartResponse,
    SimulationStopRequest,
    SimulationStopResponse,
)
from palaestrai.types.mode import Mode

from . import LOG


class SimulationSetupError(RuntimeError):
    def __init__(self, experiment_run_id, message):
        super().__init__(message)
        self.message = message
        self.experiment_run_id = experiment_run_id

    def __str__(self):
        return "%s (in experiment run: %s)" % (
            self.message,
            self.experiment_run_id,
        )


class SimulationController(ABC):
    """The base class for simulation controllers.

    A :class:`.SimulationController` coordinates the core
    between agents and environment(s). Each simulation controller is
    responsible for one simulation run. The run_id identifies the
    simulation controller.

    Parameters
    ----------
    rungov_connection : str
        A tcp string defining the location of the corresponding run
        governor. The string has the form:

            "tcp://127.0.0.1:5555"

    sim_connection : str
        A tcp string defining the location of this simulation
        controller.
    agent_conductor_ids : list
        A list of :class:`uuid.UUID`s of the agent conductors of this
        run.
    environment_conductor_ids : list
        A list of :class:`uuid.UUID`s of the environment conductors of
        this run.
    agents_config : dict
        Dictionary of agent configs
    termination_conditions : list
        A list of dicts containing import strings and params for
        termination conditions.
    mode : Mode
        Defines the mode of the run, e.g. train or test

    """

    def __init__(
        self,
        rungov_connection,
        sim_connection,
        agent_conductor_ids,
        environment_conductor_ids,
        agents_config,
        termination_conditions,
        mode,
    ):
        self._uid = str(uuid.uuid4())
        self.experiment_run_id = None
        self.mode: Mode = mode

        self._run_governor_uri = rungov_connection
        self.sim_connection = sim_connection
        self.agent_configs = agents_config
        self.ac_ids = agent_conductor_ids
        self.ec_ids = environment_conductor_ids
        LOG.debug(
            "SimulationController loading TerminationCondition objects %s.",
            termination_conditions,
        )
        try:
            self.termination_conditions = [
                _load_with_params(cond["name"], cond["params"])
                for cond in termination_conditions
            ]
        except Exception as e:
            LOG.fatal("Could not load termination condition: %s.", e)
            raise e
        LOG.debug(
            "Create %s(%s) for experiment_run_id=%s "
            "with termination_condition=%s.",
            self.__class__,
            self.uid,
            self.experiment_run_id,
            self.termination_conditions,
        )

        self._ctx = None
        self._task = None
        self._worker = None
        self._client = None
        self._rg_client = None
        self.ac_socket = None
        self.ec_socket = None
        self.rg_id = None
        self.experiment_id = None
        self.env_ids = {}

        self.all_sensors = []
        self.all_actuators = []
        self.agents = []
        self.ex_termination = False

    @property
    def uid(self):
        return self._uid

    @property
    def client(self):
        if self._client is None:
            self._client = MajorDomoClient(self.sim_connection)
        return self._client

    @property
    def worker(self):
        if self._worker is None:
            self._worker = MajorDomoWorker(
                broker_uri=self._run_governor_uri,
                service=self.experiment_run_id,
            )
        return self._worker

    @property
    def run_gov_client(self):
        if self._rg_client is None:
            self._rg_client = MajorDomoClient(self._run_governor_uri)
        return self._rg_client

    @abstractmethod
    async def simulation(self):
        pass

    async def conductor_shutdown(self):
        """Function to shutdown the conductors

        One part of the shutdown procedure is the conductor
        shutdown. This function shuts down all agent and
        environment conductors.

        """
        LOG.info(
            "SimulationController %s(id=0x%x, uid=%s) shutting down "
            "conductors.",
            self.__class__,
            id(self),
            self.uid,
        )
        msg = ShutdownRequest(self.experiment_run_id)
        for acuid in self.ac_ids:
            try:
                _ = await self.client.send(acuid, msg)
            except TypeError:
                LOG.info(
                    "SimulationController(id=0x%x, uid=%s): "
                    "AgentConductor shutdown done.",
                    id(self),
                    self.uid,
                )
        for ecuid in self.ec_ids:
            try:
                _ = await self.client.send(ecuid, msg)
            except TypeError:
                LOG.info(
                    "SimulationController(id=0x%x, uid=%s): "
                    "EnvironmentConductor shutdown done.",
                    id(self),
                    self.uid,
                )

    async def agent_shutdown(self, complete_shutdown):
        """Function to shutdown an agent

        This function is shutting down an agent. An agent is in this case
        a muscle. The parameter complete_shutdown indicates if the brain
        should also shut down. This is the case if the experiment has
        finished.

        Parameters
        ----------
        complete_shutdown : bool
            Indicates if the experiment has finished or if only the
            muscle has to shut down.
        """
        LOG.info(
            "SimulationController %s(id=0x%x, uid=%s) shutting down "
            "agents.",
            self.__class__,
            id(self),
            self.uid,
        )
        agents_to_remove = []
        while self.agents:
            LOG.debug(
                "SimulationController %s(id=0x%x, uid=%s) has %d agents "
                "to stop.",
                self.__class__,
                id(self),
                self.uid,
                len(self.agents),
            )

            for agent in self.agents:
                msg = AgentShutdownRequest(
                    self.experiment_run_id, str(agent.uid), complete_shutdown
                )
                LOG.debug(
                    "SimulationController %s(id=0x%x, uid=%s) sending "
                    "AgentShutdownRequest(agent.uid=%s).",
                    self.__class__,
                    id(self),
                    self.uid,
                    agent.uid,
                )

                response = await self.client.send(agent.uid, msg)
                if isinstance(response, AgentShutdownResponse):
                    agents_to_remove.append(agent)
                elif response is None:
                    pass

            for agent in agents_to_remove:
                self.agents.remove(agent)
            agents_to_remove = None

    async def env_shutdown(self):
        """Function to shutdown the environment

        This functions shuts down an environment. Currently
        it shuts down all available environments. Partial
        shutdown is not possible.

        """
        LOG.info(
            "SimulationController %s(id=0x%x, uid=%s) shutting down "
            "environments.",
            self.__class__,
            id(self),
            self.uid,
        )
        envs_to_remove = []
        while self.env_ids:
            for env in self.env_ids:
                msg = EnvironmentShutdownRequest(
                    self.experiment_run_id, str(env)
                )
                response = await self.client.send(env, msg)
                if isinstance(response, EnvironmentShutdownResponse):
                    envs_to_remove.append(env)
                elif response is None:
                    pass

            LOG.debug(
                "SimulationController %s(id=0x%x, uid=%s) "
                "removing environments %s from the internal list: %s",
                self.__class__,
                id(self),
                self.uid,
                envs_to_remove,
                self.env_ids,
            )
            for env in envs_to_remove:
                _ = self.env_ids.pop(env, None)
            envs_to_remove = None

    async def stop_simulation(self, complete_shutdown):
        """Coordinating the shutdown

        This function coordinates the shutdown of the individual components.

        Parameters
        ----------
        complete_shutdown : bool
            Indicates if the experiment has finished or if only the
            muscle has to shut down.
        """
        LOG.info(
            "SimulationController %s(id=0x%x, uid=%s) stopping simulation.",
            self.__class__,
            id(self),
            self.uid,
        )
        await self.agent_shutdown(complete_shutdown)
        await self.env_shutdown()
        await self.conductor_shutdown()

    def _setup(self):
        LOG.debug(
            "SimulationController %s(id=0x%x, uid=%s) creating new "
            "MajorDomoWorker for experiment_run_id=%s.",
            self.__class__,
            id(self),
            self.uid,
            self.experiment_run_id,
        )
        self._worker = MajorDomoWorker(
            broker_uri=self._run_governor_uri, service=self.uid
        )

    def _generate_uid(self, gen_type):
        if gen_type == "agent":
            uid = self.experiment_run_id + "_A:" + str(uuid.uuid4())
        elif gen_type == "environment":
            uid = self.experiment_run_id + "_E:" + str(uuid.uuid4())
        else:
            uid = uuid.uuid4()
        return uid

    async def _init_simulation(self):
        LOG.info(
            "SimulationController %s(id=0x%x, uid=%s) "
            "starting environment conductors.",
            self.__class__,
            id(self),
            self.uid,
        )
        for ec in self.ec_ids:
            uid = self._generate_uid("environment")
            self.env_ids[uid] = ec
            LOG.debug(
                "SimulationController %s(id=0x%x, uid=%s) "
                "sending EnvironmentSetupRequest(experiment_run_id=%s, "
                "environment_id=%s, environment_conductor_id=%s).",
                self.__class__,
                id(self),
                self.uid,
                self.experiment_run_id,
                uid,
                ec,
            )
            request = EnvironmentSetupRequest(
                environment_id=uid,
                experiment_run_id=self.experiment_run_id,
                receiver_environment_conductor_id=ec,
                sender_simulation_controller_id=self.uid,
            )
            response = await self.client.send(ec, request)
            if not response:
                LOG.critical(
                    "SimulationController %s(id=0x%x, uid=%s) "
                    "received no EnvironmentSetupResponse after "
                    "sending EnvironmentSetupRequest(experiment_run_id=%s, "
                    "environment_id=%s, environment_conductor_id=%s).",
                    self.__class__,
                    id(self),
                    self.uid,
                    self.experiment_run_id,
                    uid,
                    ec,
                )
                raise SimulationSetupError(
                    experiment_run_id=self.experiment_run_id,
                    message="environment %s did not respond "
                    "to EnvironmentSetupRequest." % ec,
                )

            request = EnvironmentStartRequest(self.experiment_run_id, uid)
            response = await self.client.send(uid, request)

            if not response:
                LOG.critical(
                    "SimulationController %s(id=0x%x, uid=%s) "
                    "received no EnvironmentStartResponse after "
                    "sending EnvironmentStartRequest(experiment_run_id=%s, "
                    "environment_id=%s, environment_conductor_id=%s).",
                    self.__class__,
                    id(self),
                    self.uid,
                    self.experiment_run_id,
                    uid,
                    ec,
                )
                raise SimulationSetupError(
                    experiment_run_id=self.experiment_run_id,
                    message="environment %s did not respond "
                    "to EnvironmentStartRequest" % ec,
                )
            if isinstance(response, EnvironmentStartResponse):
                self.all_sensors.extend(response.sensors)
                self.all_actuators.extend(response.actuators)
        LOG.info(
            "SimulationController %s(id=0x%x, uid=%s) "
            "starting agent conductors.",
            self.__class__,
            id(self),
            self.uid,
        )
        for acuid in self.ac_ids:
            conf = self.agent_configs[acuid]
            uid = self._generate_uid("agent")

            sensors, actuators = self.access_list(
                conf["sensors"], conf["actuators"]
            )
            agent = Agent(
                uid,
                conf["brain"]["name"],
                conf["brain"]["params"],
                conf["muscle"]["name"],
                conf["muscle"]["params"],
                sensors,
                actuators,
            )
            LOG.debug("Agents' sensors: %s", agent.sensors)
            self.agents.append(agent)
            LOG.debug(
                "SimulationController %s(id=0x%x, uid=%s) sending "
                "AgentSetupRequest(agent_conductor_id=%s, agent_id=%s).",
                self.__class__,
                id(self),
                self.uid,
                acuid,
                uid,
            )
            request = AgentSetupRequest(
                sender_simulation_controller=self.uid,
                receiver_agent_conductor=str(acuid),
                experiment_run_id=self.experiment_run_id,
                agent_id=uid,
                agent_name=conf["name"] if "name" in conf else None,
                sensors=sensors,
                actuators=actuators,
            )

            response = await self.client.send(acuid, request)
            if not response:
                LOG.critical(
                    "SimulationController %s(id=0x%x, uid=%s) "
                    "received no AgentSetupResponse after "
                    "sending AgentSetupRequest(receiver=%s, "
                    "experiment_run_id=%s, agent_id=%s, agent_name=%s).",
                    self.__class__,
                    id(self),
                    self.uid,
                    request.receiver_agent_conductor,
                    self.experiment_run_id,
                    uid,
                    conf["name"] if "name" in conf else None,
                )
                raise SimulationSetupError(
                    experiment_run_id=self.experiment_run_id,
                    message="AgentConductor %s did not respond "
                    "to AgentSetupRequest" % request.receiver,
                )

    async def run(self):
        """Main function, coordination of messages

        This function takes care of all the coordination. It performs
        the initial setup, it receives new messages and processes these
        messages. In addition, it also takes care of the shutdown.
        """
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGABRT, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        setproctitle.setproctitle(
            "palaestrAI[SimulationController-%s]" % self._uid[-6:]
        )
        LOG.info(
            "SimulationController %s(id=0x%x, uid=%s) standing by for "
            "experiment_run_id=%s.",
            self.__class__,
            id(self),
            self.uid,
            self.experiment_run_id,
        )
        self._setup()
        reply = None
        while not self._task or not self._task.done():
            LOG.debug(
                "SimulationController %s(id=0x%x, uid=%s) waiting for "
                "next request.",
                self.__class__,
                id(self),
                self.uid,
            )

            transceive_task = asyncio.create_task(
                self.worker.transceive(reply)
            )
            tasks_done, _ = await asyncio.wait(
                {self._task, transceive_task}
                if self._task
                else {transceive_task},
                return_when=asyncio.FIRST_COMPLETED,
            )
            if transceive_task in tasks_done:
                request = transceive_task.result()
            if self._task in tasks_done:
                continue

            LOG.info(
                "SimulationController %s(id=0x%x, uid=%s) received "
                "request: %s.",
                self.__class__,
                id(self),
                self.uid,
                request,
            )

            if isinstance(request, SimulationStartRequest):
                LOG.debug(
                    "SimulationController %s(id=0x%x, uid=%s) received "
                    "SimulationStartRequest from RunGovernor(%s) for "
                    "experiment_run_id=%s.",
                    self.__class__,
                    id(self),
                    self.uid,
                    request.run_governor_id,
                    self.experiment_run_id,
                )
                self.rg_id = request.run_governor_id
                self.experiment_run_id = request.experiment_run_id

                reply = SimulationStartResponse()
                await self._init_simulation()
                self._task = asyncio.create_task(self.simulation())

                LOG.debug(
                    "SimulationController %s(id=0x%x, uid=%s) "
                    "initialized simulation.",
                    self.__class__,
                    id(self),
                    self.uid,
                )
            elif isinstance(request, SimulationStopRequest):
                LOG.debug(
                    "SimulationController %s(id=0x%x, uid=%s) received "
                    "SimulationStopRequest(experiment_run_id=%s).",
                    self.__class__,
                    id(self),
                    self.uid,
                    self.experiment_run_id,
                )
                self.ex_termination = True

                reply = SimulationStopResponse(self.experiment_run_id)
                # self._task.cancel()  # We probably don't need this
                try:
                    # Maybe add a timeout here and cancel once we're
                    # timed out.
                    await self._task
                except asyncio.CancelledError:
                    pass  # We actually expect this, more or less.

                LOG.debug(
                    "SimulationController %s(id=0x%x, uid=%s) finished "
                    "simulation of ExperimentRun(id=%s), sending "
                    "SimulationStopResponse(experiment_run_id=%s).",
                    self.__class__,
                    id(self),
                    self.uid,
                    self.experiment_run_id,
                    self.experiment_run_id,
                )

        LOG.debug(
            "SimulationController %s(id=0x%x, uid=%s) sending final reply.",
            self.__class__,
            id(self),
            self.uid,
        )
        await self.worker.transceive(reply, skip_recv=True)
        LOG.info(
            "SimulationController %s(id=0x%x, uid=%s) completes shutdown.",
            self.__class__,
            id(self),
            self.uid,
        )

    @abstractmethod
    def access_list(self, sensor_list, actuator_list):
        pass


def _load_with_params(module_path, params):
    """Load a class from *module_path* and pass *params*."""
    module, clazz = module_path.split(":")
    module = import_module(module)
    return getattr(module, clazz)(**params)



from dataclasses import dataclass


@dataclass
class EnvironmentSetupResponse:
    """Signals successful environment setup and delivers environment parameters

    * Sender: :class:`EnvironmentConductor´
    * Receiver: :class:`SimulationController`

    :param sender_environment_conductor: ID of the sending
        :class:`EnvironmentConductor`
    :param receiver_simulation_controller: ID of the receiving
        :class:`SimulationController`
    :param experiment_run_id: ID of the experiment run
    :param environment_id: ID of the newly setup environment
    :param environment_parameters: All parameters that describe the
        environment that has just been set up
    """

    environment_id: str
    experiment_run_id: str
    sender_environment_conductor: str
    receiver_simulation_controller: str
    environment_type: str
    environment_parameters: dict

    @property
    def sender(self):
        return self.sender_environment_conductor

    @property
    def receiver(self):
        return self.receiver_simulation_controller



import itertools
from typing import TYPE_CHECKING

from palaestrai.util.exception import PrematureTaskDeathError

from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from palaestrai.experiment import RunGovernor


class RGSHandlingDeadChildren(RunGovernorState):
    """Represent the HANDLING_DEAD_CHILDREN state of the run governor.

    Possible next states are:
    * :class:`.RGSStoppingTransceiving` in the case of errors.
    * :class:`.RGSStoppingSimulation` in the default case.

    Parameters
    ----------
    rgc: :class:`.RunGovernor`
        The run governor instance that provides the context for this
        state.


    """

    def __init__(self, rgc: "RunGovernor"):
        super().__init__(rgc, "HANDLING_DEAD_CHILDREN")

    async def run(self):
        all_processes = [
            proc
            for proc in itertools.chain(
                self.rgc.sim_controllers.values(),
                self.rgc.env_conductors.values(),
                self.rgc.agent_conductors.values(),
            )
        ]
        dead_children = [proc for proc in all_processes if not proc.is_alive()]

        # Clean up: Either we have processes that ended normally, then
        # we just free ressources and reap the children. If some
        # process died from a premature error, we have to shut down
        # the whole experiment, as it will doubtedly have any value to
        # continue it.

        if dead_children:

            def _clean(d: dict):
                return {
                    uid: proc for uid, proc in d.items() if proc.is_alive()
                }

            self.rgc.agent_conductors = _clean(self.rgc.agent_conductors)
            self.rgc.env_conductors = _clean(self.rgc.env_conductors)
            self.rgc.sim_controllers = _clean(self.rgc.sim_controllers)

        premature_deaths = [
            proc for proc in dead_children if proc.exitcode != 0
        ]
        if premature_deaths:
            LOG.critical(
                "RunGovernor(id=0x%x, uid=%s) "
                "has suffered from premature task death: %s.",
                id(self.rgc),
                self.rgc.uid,
                premature_deaths,
            )
            self.add_error(PrematureTaskDeathError())

    def next_state(self):
        from . import RGSStoppingSimulation, RGSStoppingTransceiving

        if len(self.rgc.errors) > 0:
            self.rgc.state = RGSStoppingTransceiving(self.rgc)
        else:
            self.rgc.state = RGSStoppingSimulation(self.rgc)



import numpy as np

from palaestrai.types import Box
from palaestrai.types import Discrete
from palaestrai.types import MultiDiscrete
from palaestrai.types import MultiBinary
from palaestrai.types import Tuple


def flatdim(space):
    if isinstance(space, Box):
        return int(np.prod(space.shape))
    elif isinstance(space, Discrete):
        return int(space.n)
    elif isinstance(space, Tuple):
        return int(sum([flatdim(s) for s in space.spaces]))
    elif isinstance(space, MultiBinary):
        return int(space.n)
    elif isinstance(space, MultiDiscrete):
        return int(np.prod(space.shape))
    else:
        raise NotImplementedError


def flatten(space, x):
    if isinstance(space, Box):
        return np.asarray(x, dtype=np.float32).flatten()
    elif isinstance(space, Discrete):
        onehot = np.zeros(space.n, dtype=np.float32)
        onehot[x] = 1.0
        return onehot
    elif isinstance(space, Tuple):
        return np.concatenate(
            [flatten(s, x_part) for x_part, s in zip(x, space.spaces)]
        )
    elif isinstance(space, MultiBinary):
        return np.asarray(x).flatten()
    elif isinstance(space, MultiDiscrete):
        onehot = np.zeros(np.sum(space.nvec), dtype=np.float32)
        xj = 0
        for xi, si in zip(x, space.nvec):
            onehot[xi + xj] = 1.0
            xj = xj + si
        return onehot
    else:
        raise NotImplementedError


def unflatten(space, x):
    if isinstance(space, Box):
        return np.asarray(x, dtype=np.float32).reshape(space.shape)
    elif isinstance(space, Discrete):
        return int(np.nonzero(x)[0][0])
    elif isinstance(space, Tuple):
        dims = [flatdim(s) for s in space.spaces]
        list_flattened = np.split(x, np.cumsum(dims)[:-1])
        list_unflattened = [
            unflatten(s, flattened)
            for flattened, s in zip(list_flattened, space.spaces)
        ]
        return tuple(list_unflattened)
    elif isinstance(space, MultiBinary):
        return np.asarray(x).reshape(space.shape)
    elif isinstance(space, MultiDiscrete):
        result = []
        s = 0
        for si in space.nvec:
            for xi in range(s, s + si):
                if x[xi] == 1.0:
                    result.append(xi - s)
            s = s + si
        if space.nvec.shape == np.array(result).shape:
            return result
    else:
        raise NotImplementedError



from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Optional

if TYPE_CHECKING:
    from palaestrai.agent import SensorInformation, RewardInformation


@dataclass
class MuscleUpdateRequest:
    """Notifies the :class:`Brain` that a :class:`Muscle` of an action.

    * Sender: A :class:`Muscle` after acting
    * Receiver: The :class:`Brain`

    :param sender_muscle_id: ID of the sending :class:`Muscle`
    :param receiver_brain_id: ID of the receiving :class:`Brain`
    :param experiment_run_id: ID of the experiment run this action happened in
    :param agent_id: ID of the :class:`Agent` muscle and brain belong to
    :param sensor_readings: A list of sensor readings on which the muscle
        acted
    :param last_actions: A list of actions the muscle proposed to do last
    :param reward: Reward received from the last action
    :param is_terminal: Indicates whether this was the last action as the
        environment (or agent) are done
    :param shutdown: Indiciates system shutdown
    """

    sender_muscle_id: str
    receiver_brain_id: Optional[str]
    experiment_run_id: str
    agent_id: str
    sensor_readings: List[SensorInformation]
    last_actions: List
    reward: List[RewardInformation]
    is_terminal: bool
    shutdown: bool

    @property
    def sender(self):
        return self.sender_muscle_id

    @property
    def receiver(self):
        return self.receiver_brain_id



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class ExperimentRunShutdownRequest:
    """Request to a :class:`RunGovernor` to stop an experiment run.

    Parameters
    ----------
    sender_executor_id: str
        ID of the sending executor.
    receiver_run_governor_id: str
        ID of the receiving run governor.
    experiment_run: :class:`ExperimentRun`
        The experiment run that should be executed.

    """

    sender_executor_id: str
    receiver_run_governor_id: str
    experiment_run_id: str

    @property
    def sender(self):
        """ID of the sending executor."""
        return self.sender_executor_id

    @property
    def receiver(self):
        """ID of the receiving run governor."""
        return self.receiver_run_governor_id



from __future__ import annotations
from dataclasses import dataclass

# from typing import TYPE_CHECKING


@dataclass
class MuscleShutdownRequest:
    """Notifies the :class:`.Brain` that the :class:`.Muscle received
    a shutdown request.

    Parameters
    ----------
    sender_muscle_id: str
        The uid of the sender muscle.
    agent_id: str
        The uid of the corresponding :class:`.Agent`.
    experiment_run_id: str
        The uid of the corresponding experiment run.

    """

    sender_muscle_id: str
    agent_id: str
    experiment_run_id: str

    @property
    def sender(self):
        return self.sender_muscle_id



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class NextPhaseResponse:
    """A response of a :class:`RunGovernor` to itself for a next phase
    request.

    This message object might seem a bit unnecessary at first. But to
    not (further) violate the major domo pattern, the run governor
    needs to send a reply in the transceiving state. But unless the run
    governor has received an ExperimentRunStartRequest or an
    ExperimentRunShutdownRequest, no reply is available.

    This response is send although there is no one waiting for it.

    Parameters
    ----------
    sender_run_governor_id: str
        ID of the sending run governor
    receiver_run_governor_id: str
        ID of the receiving run governer (it's the same as the sender).
    has_next_phase: bool
        True if at least one phase is following.
    """

    sender_run_governor_id: str
    receiver_run_governor_id: str
    has_next_phase: bool

    @property
    def sender(self):
        """ID of the sending run governor."""
        return self.sender_run_governor_id

    @property
    def receiver(self):
        """ID of the receiving run governor."""
        return self.receiver_run_governor_id



import io
import uuid
import unittest
import numpy as np
from unittest import TestCase
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

import palaestrai.store.database_model as model


class TestDatabaseModel(TestCase):
    _URI = "sqlite://"

    def setUp(self):
        self._engine = create_engine(TestDatabaseModel._URI)
        model.Base.metadata.bind = self._engine
        self._Session = sessionmaker(bind=self._engine)
        self._session = self._Session()

    def tearDown(self) -> None:
        self.store.close()

    @property
    def store(self):
        return self._session

    def test_create(self):
        try:
            model.Base.metadata.create_all()
        except Exception as e:
            self.fail(
                "model.Base.metadata.create_all() must not raise: %s" % e
            )

    def test_add(self):
        """
        Tests addition with all relevant relationships, confirming
        the schema
        """
        model.Base.metadata.create_all()

        experiment_study_entity = model.ExperimentStudy(
            name="Test", study="many things"
        )
        experiment_1 = model.Experiment(experiment="many more things")
        experiment_2 = model.Experiment(experiment="many more things again")

        experiment_study_entity.experiments.append(experiment_1)
        experiment_study_entity.experiments.append(experiment_2)

        experiment_run_entity = model.ExperimentRun(uuid=str(uuid.uuid4()))
        experiment_1.experiment_runs.append(experiment_run_entity)

        agent_conductor_entity = model.AgentConductor(
            uuid="Momzobo", name="the greatest conductor"
        )
        experiment_run_entity.agent_conductors.append(agent_conductor_entity)

        brain_entity = model.Brain(parameters=np.random.rand(3, 7))
        agent_conductor_entity.brain.append(brain_entity)

        brain_config_entity = model.BrainConfiguration(
            agent_id="0815", simtime=1387, parameters=np.random.rand(2, 4)
        )
        brain_entity.brain_configs.append(brain_config_entity)

        simulation_instance_entity = model.SimulationInstance(
            uuid=str(uuid.uuid4())
        )
        experiment_run_entity.simulation_instances.append(
            simulation_instance_entity
        )

        environment_conductor_entity = model.EnvironmentConductor(
            uuid="Test Environment Conductor"
        )
        simulation_instance_entity.environment_conductors.append(
            environment_conductor_entity
        )

        muscle_entity = model.Muscle(uuid="MuscleRoaar")
        agent_conductor_entity.muscles.append(muscle_entity)
        simulation_instance_entity.muscles.append(muscle_entity)

        muscle_states_entity = model.MuscleState(
            parameters=np.random.rand(3, 2)
        )
        muscle_entity.muscle_states.append(muscle_states_entity)

        muscle_actions_entity = model.MuscleAction(
            simtime=19786,
            actuator_information=np.zeros((2, 7)),
        )
        muscle_entity.muscle_actions.append(muscle_actions_entity)

        muscle_sensor_readings_entity = model.MuscleSensorReading(
            active_agent_id=42,
            sensor_id=13,
            simtime=19786,
            value=np.random.rand(7),
        )
        muscle_entity.muscle_sensor_readings.append(
            muscle_sensor_readings_entity
        )

        world_states_entity = model.WorldState(
            simtime=19786,
            state_dump=np.random.rand(5, 7),
        )
        environment_conductor_entity.world_states.append(world_states_entity)
        self.store.add(world_states_entity)
        self.store.commit()
        try:
            run_e = self.store.query(model.ExperimentRun).one()
        except Exception as e:
            self.fail("Query.one must succeed, but raised: %s" % e)

    def test_query_by_experiment_run_id(self):
        self.test_add()
        try:
            run_e: model.ExperimentRun = self.store.query(
                model.ExperimentRun
            ).one()
        except Exception as e:
            self.fail("Query.one must succeed, but raised: %s" % e)
        ety = (
            self.store.query(
                model.ExperimentRun,
                model.SimulationInstance,
                model.MuscleState,
            )
            .join(model.SimulationInstance)
            .join(model.Muscle)
            .join(model.MuscleState)
        )
        try:
            param = ety.one()[2].parameters
        except Exception as e:
            self.fail("Query.join.one must succeed, but raised: %s" % e)
        self.assertEqual(np.frombuffer(param).shape, (6,))


if __name__ == "__main__":
    unittest.main()



class AgentConductorSetupResponse:
    def __init__(self, run_id, conductor_id):
        self._run_id = run_id
        self._agent_conductor_id = conductor_id

    @property
    def run_id(self):
        return self._run_id

    @property
    def agent_conductor_id(self):
        return self._agent_conductor_id



import unittest

from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import (
    RGSErrorHandlingInitializing,
    RGSStoppingTransceiving,
)


class TestHandlingErrorsInitializing(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)
        self.rgs = RGSErrorHandlingInitializing(self.rungov)
        self.rungov.state = self.rgs

    async def test_run(self):
        logger = f"{RunGovernor.__module__}".rsplit(".", 1)[0]

        with self.assertLogs(logger, level="DEBUG") as cm:
            await self.rgs.run()

        self.assertIn("Handling these errors now.", cm.output[0])

    def test_next_state_success(self):

        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSStoppingTransceiving)



import unittest
from unittest.mock import MagicMock

from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import (
    RGSErrorHandlingTransceiving,
    RGSHandlingDeadChildren,
    RGSStoppingSimulation,
    RGSTransceiving,
)
from palaestrai.util.exception import (
    DeadChildrenRisingAsZombiesError,
    InvalidRequestError,
    RequestIsNoneError,
    SignalInterruptError,
    TasksNotFinishedError,
)


class TestHandlingErrorsStarting(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)
        self.rgs = RGSErrorHandlingTransceiving(self.rungov)
        self.rungov.state = self.rgs

    async def test_run(self):
        logger = f"{RunGovernor.__module__}".rsplit(".", 1)[0]

        self.rgs.add_error(TasksNotFinishedError())

        with self.assertLogs(logger, level="DEBUG") as cm:
            await self.rgs.run()

        self.assertIn("Handling these errors now.", cm.output[0])

    def test_next_state_invalid_request(self):

        self.rgs.add_error(InvalidRequestError(MagicMock(), MagicMock()))
        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSTransceiving)

    def test_next_state_task_not_finished(self):

        self.rgs.add_error(TasksNotFinishedError())
        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSTransceiving)

    def test_next_state_signal_interrupt_error(self):
        self.rgs.add_error(SignalInterruptError())
        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSStoppingSimulation)

    def test_next_state_dead_children(self):

        self.rgs.add_error(DeadChildrenRisingAsZombiesError())
        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSHandlingDeadChildren)

    def test_next_state_request_is_none(self):

        self.rgs.add_error(RequestIsNoneError())
        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSTransceiving)



""""Majordomo Protocol Client API, Python version.
Implements the MDP/Worker spec at http:#rfc.zeromq.org/spec:7.
Author: Min RK <benjaminrk@gmail.com>
Based on Java example by Arkadiusz Orzechowski
"""

import zmq
import zmq.asyncio

# from .zhelpers import pprint_message
from palaestrai.core.serialisation import deserialize, serialize

from . import LOG, RuntimeConfig
from . import MDP
from .zhelpers import ContextGuard


class MajorDomoClient:
    """Client object for distributing tasks to workers and receiving results

    The major domo protocol client is the initiator of tasks. It sends
    messages that are distributed according to a service name to worksers. The
    main method of this class, :py:func:`MajorDomoClient.send`, sends such
    a request and waits for the reply of the corresponding worker.

    Major domo clients strictly adhere to the request-response pattern:
    A client always awaits for a reply when sending a request.

    Majordomo Protocol Client API, Python version.
    Implements the MDP/Worker spec at http:#rfc.zeromq.org/spec:7.
    """

    def __init__(self, broker_uri: str):
        self._socket = None
        self._broker_uri = broker_uri
        self._poller = zmq.asyncio.Poller()
        self.timeout = RuntimeConfig().major_domo_client_timeout
        self.retries = RuntimeConfig().major_domo_client_retries

    def reconnect_to_broker(self):
        """Connect or reconnect to broker

        Attempts to connect to the broker. This method can be called multiple
        times: For initial connection, or for reconnection.
        """

        if self._socket:
            self._poller.unregister(self._socket)
            self._socket.close()
        self._socket = zmq.asyncio.Socket(
            context=ContextGuard.asyncio_context(), socket_type=zmq.REQ
        )  # self.ctx.socket(zmq.REQ)
        self._socket.linger = 0
        self._socket.connect(self._broker_uri)
        self._poller.register(self._socket, zmq.POLLIN)
        LOG.debug(
            "MajorDomoClient(id=%s) connected to MajorDomoBroker(uri=%s)",
            id(self),
            self._broker_uri,
        )

    async def destroy(self):
        if not self._socket:
            self.reconnect_to_broker()
        request = [MDP.C_CLIENT, MDP.W_DESTROY]
        LOG.debug(
            "MajorDomoClient(id=%s) sending destroy message "
            "to MajorDomoBroker(uri=%s)",
            id(self),
            self._broker_uri,
        )
        await self._socket.send_multipart(request)

    async def send(self, service, request):
        """Send request to broker and get reply by hook or crook.

        Takes ownership of request message and destroys it when sent.
        Returns the reply message or None if there was no reply.
        """
        assert not isinstance(request, bytes)
        request = serialize(request)

        if not isinstance(service, bytes):
            service = bytes(str(service), "ascii")

        if not self._socket:
            self.reconnect_to_broker()

        if not isinstance(request, list):
            request = [request]
        request = [MDP.C_CLIENT, service] + request

        retries = self.retries
        while retries > 0:
            LOG.debug(
                "MajorDomoClient(id=%s) sending request to service '%s'",
                id(self),
                service,
            )
            await self._socket.send_multipart(request)
            try:
                items = await self._poller.poll(self.timeout)
            except KeyboardInterrupt:
                return None
            except SystemExit:
                return None
            if items:
                msg = await self._socket.recv_multipart()
                LOG.debug("MajorDomoClient(id=%s) received reply", id(self))

                if len(msg) < 3:
                    LOG.error(
                        "MajorDomoClient(id=0x%x) received message "
                        "of length = %d, should be >= 3; ignoring",
                        id(self),
                        len(msg),
                    )
                    break
                header = msg.pop(0)
                if header != MDP.C_CLIENT:
                    LOG.error(
                        "MajorDomoClient(id=0x%x) received message "
                        "with header type '%s', but expected '%s'; ignoring",
                        id(self),
                        header,
                        MDP.C_CLIENT,
                    )
                    break

                reply_service = msg.pop(0)
                if service != reply_service:
                    LOG.error(
                        "MajorDomoClient(id=0x%x) received reply "
                        "for service '%s', but waited for service '%s'; "
                        "returning for the caller to enjoy the unwanted gift",
                        id(self),
                        service,
                        reply_service,
                    )
                return deserialize(msg)
            else:
                if retries:
                    LOG.warning(
                        "MajorDomoClient(id=%s) "
                        "received no response from service '%s' "
                        "for message %s "
                        "after %s of %s attempt(s) with timeout %s, "
                        "reconnecting and retrying",
                        id(self),
                        service,
                        str(request),
                        self.retries - retries,
                        self.retries,
                        self.timeout,
                    )
                    self.reconnect_to_broker()
                else:
                    LOG.error(
                        "MajorDomoClient(id=0x%x) suffering from permanent "
                        "connection error: Tried service '%s' %s times, "
                        "but got no response for request %s, "
                        "abondoning this message",
                        id(self),
                        service,
                        self.retries,
                        str(request),
                    )
                    break
                retries -= 1
        return None



#!usr/bin/env python3
# manager.py -- Manager CLI for ARL execution with palaestrAI
# Copyright (C) 2020  OFFIS, All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.

import logging
import logging.config
import sys

import click
import pkg_resources
import setproctitle

import palaestrai.core.runtime_config
from palaestrai.cli import client_local
from palaestrai.core import RuntimeConfig


@click.group()
@click.option(
    "-c",
    "--config",
    type=click.Path(),
    help="Supply custom runtime configuration file. "
    "(Default search path: %s)"
    % (palaestrai.core.runtime_config._RuntimeConfig.CONFIG_FILE_PATHS),
)
@click.option(
    "--verbose",
    "-v",
    count=True,
    help="Increases the program verbosity, can be given numerous times: "
    "-v prints also INFO messages, and -vv emits DEBUG output."
    "output",
)
@click.version_option(pkg_resources.require("palaestrai")[0].version)
def cli(config=None, verbose=0):
    setproctitle.setproctitle(" ".join(sys.argv))

    if config:
        try:
            with open(config, "r") as fp:
                RuntimeConfig().load(fp)
        except OSError as e:
            click.echo(
                "ERROR: Could not load config from %s: %s." % (config, e),
                file=sys.stderr,
            )
            exit(1)
    else:
        try:
            RuntimeConfig().load()
        except FileNotFoundError as e:
            click.echo(
                "Please create a runtime config. %s.\n"
                "Will continue with built-in defaults." % e,
                file=sys.stderr,
            )
    init_logger(verbose)


def init_logger(verbose):
    """Init logger with config from either RuntimeConfig or a default."""
    levels = [logging.WARNING, logging.INFO, logging.DEBUG]
    log_level = levels[verbose if verbose < len(levels) else len(levels) - 1]

    try:
        logging.config.dictConfig(RuntimeConfig().logging)
        logging.debug(
            "Initialized logging from RuntimeConfig(%s)", RuntimeConfig()
        )
    except (KeyError, ValueError) as e:
        logging.basicConfig(level=log_level)
        logging.warning(
            "Could not load logging config (%s), continuing with defaults",
            e,
        )
    if verbose != 0:
        for name in logging.root.manager.loggerDict:
            logging.getLogger(name).setLevel(log_level)
            RuntimeConfig().logging["loggers"].update(
                {name: {"level": str(logging._levelToName[log_level])}}
            )


for cmd in [
    client_local.experiment_start,
    client_local.database_create,
    client_local.database_migrate,
    client_local.runtime_config_show_default,
    client_local.runtime_config_show_effective,
]:
    cli.add_command(cmd)


if __name__ == "__main__":
    cli()



from __future__ import annotations
from dataclasses import dataclass
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from palaestrai.agent import SensorInformation
    from palaestrai.agent import RewardInformation


@dataclass
class EnvironmentUpdateResponse:
    """Reports the current state of the environment.

    * Sender: :class:`Environment`
    * Receiver: :class:`SimulationController

    :param sender_environment_id: ID of the sending :class:`Environment`
    :param receiver_simulation_controller_id: ID of the receiving
        :class:`SimulationController`
    :param experiment_run_id: ID of the current experiment run the environment
        participates in
    :param environment_conductor_id: ID of the :class:`EnvironmentConductor`
        the environment belongs to
    :param sensors: Current list of sensor data
    :param reard: Reward given by the environment
    :param is_terminal: Indicates whether the environment has reached
        a terminal state
    """

    sender_environment_id: str
    receiver_simulation_controller_id: str
    experiment_run_id: str
    environment_conductor_id: str
    sensors: List[SensorInformation]
    reward: List[RewardInformation]
    is_terminal: bool

    @property
    def sender(self):
        return self.sender_environment_id

    @property
    def receiver(self):
        return self.receiver_simulation_controller_id



import unittest
from unittest.mock import patch, MagicMock

from palaestrai.experiment import RunGovernor

from palaestrai.experiment.rungovernor_states import (
    RGSPristine,
    RGSInitializing,
)


class TestPristine(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)

    @patch(f"{RGSPristine.__module__}.os")
    async def test_run(self, mock_os):
        rgs = RGSPristine(self.rungov)
        mock_os.setpgrp = MagicMock()
        mock_os.getpgid = MagicMock()
        mock_os.getpid = MagicMock()

        await rgs.run()

        mock_os.setpgrp.assert_called_once()
        mock_os.getpgid.assert_called_once()
        mock_os.getpid.assert_called_once()

    def test_next_state(self):
        rgs = RGSPristine(self.rungov)

        rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSInitializing)


if __name__ == "__main__":
    unittest.main()



import logging

LOG = logging.getLogger(__name__)

from .agent_conductor_setup_req import AgentConductorSetupRequest
from .agent_conductor_setup_rsp import AgentConductorSetupResponse
from .agent_setup_req import AgentSetupRequest
from .agent_setup_rsp import AgentSetupResponse
from .agent_shutdown_req import AgentShutdownRequest
from .agent_shutdown_rsp import AgentShutdownResponse
from .agent_update_req import AgentUpdateRequest
from .agent_update_rsp import AgentUpdateResponse
from .environment_reset_notification_req import (
    EnvironmentResetNotificationRequest,
)
from .environment_reset_notification_rsp import (
    EnvironmentResetNotificationResponse,
)
from .environment_reset_req import EnvironmentResetRequest
from .environment_reset_rsp import EnvironmentResetResponse
from .environment_setup_req import EnvironmentSetupRequest
from .environment_setup_rsp import EnvironmentSetupResponse
from .environment_shutdown_req import EnvironmentShutdownRequest
from .environment_shutdown_rsp import EnvironmentShutdownResponse
from .environment_start_req import EnvironmentStartRequest
from .environment_start_rsp import EnvironmentStartResponse
from .environment_update_req import EnvironmentUpdateRequest
from .environment_update_rsp import EnvironmentUpdateResponse
from .run_shutdown_req import ExperimentRunShutdownRequest
from .run_shutdown_rsp import ExperimentRunShutdownResponse
from .run_start_req import ExperimentRunStartRequest
from .run_start_rsp import ExperimentRunStartResponse
from .muscle_shutdown_req import MuscleShutdownRequest
from .muscle_shutdown_rsp import MuscleShutdownResponse
from .muscle_update_req import MuscleUpdateRequest
from .muscle_update_rsp import MuscleUpdateResponse
from .shutdown_req import ShutdownRequest
from .shutdown_rsp import ShutdownResponse
from .simulation_controller_termination_req import (
    SimulationControllerTerminationRequest,
)
from .simulation_controller_termination_rsp import (
    SimulationControllerTerminationResponse,
)
from .simulation_start_req import SimulationStartRequest
from .simulation_start_rsp import SimulationStartResponse
from .simulation_stop_req import SimulationStopRequest
from .simulation_stop_rsp import SimulationStopResponse
from .next_phase_req import NextPhaseRequest
from .next_phase_rsp import NextPhaseResponse



from unittest import TestCase

from numpy.random import RandomState

from palaestrai.util.seeding import (
    create_seed,
    np_random,
    hash_seed,
    _bigint_from_bytes,
    _int_list_from_bigint,
)


class SeedingTest(TestCase):
    def test_np_random_runtime_error(self):
        # Check Runtime Errors
        self.assertRaises(RuntimeError, np_random, 12.0)
        self.assertRaises(RuntimeError, np_random, "asdf")

    def test_np_random_keep_seed(self):
        rng, seed = np_random(4)
        self.assertEqual(seed, 4)
        self.assertIsInstance(rng, RandomState)

    def test_np_random_deterministic_seed(self):
        rng, seed = np_random(4)
        self.assertEqual(seed, 4)
        self.assertIsInstance(rng, RandomState)
        rng2, seed2 = np_random()
        self.assertIsInstance(rng2, RandomState)
        self.assertIsInstance(seed2, int)

    def test_np_random_deterministic_seed2(self):
        rng, seed = np_random(4)
        rng_2, seed_2 = np_random(4)
        self.assertEqual(seed, seed_2)
        rng_state = rng._bit_generator.state
        rng2_state = rng_2._bit_generator.state
        self.assertTrue(str(rng_state) == str(rng2_state))

    def test_hash_seed(self):
        seed_1 = hash_seed()
        self.assertIsInstance(seed_1, int)
        seed_2 = hash_seed(5)
        self.assertNotEqual(seed_2, 5)
        seed_3 = hash_seed(5)
        self.assertEqual(seed_3, seed_2)

    def test_create_seed(self):
        no_seed = create_seed()
        assert isinstance(no_seed, int)

        str_seed = create_seed("9")
        assert isinstance(str_seed, int)
        str_seed2 = create_seed("9")
        assert str_seed2 == str_seed
        str_seed_diff = create_seed("asdf")
        str_seed_diff2 = create_seed("asdf")
        assert str_seed_diff != str_seed
        assert str_seed_diff2 == str_seed_diff

        int_types_seed = create_seed(9)
        int_types_seed2 = create_seed(9)
        assert isinstance(int_types_seed, int)
        assert int_types_seed == str_seed
        assert int_types_seed == int_types_seed2

    def test_bigint_from_bytes(self):
        bytes_s = b"asdfasdfdsfsdaf"
        self.assertEqual(
            _bigint_from_bytes(bytes_s), _bigint_from_bytes(bytes_s)
        )

    def test_int_list_from_bigint(self):
        int_list = _int_list_from_bigint(249213748921473982147893214)
        self.assertEqual(int_list, [1275107294, 1716681061, 13509904])



from __future__ import annotations
from dataclasses import dataclass
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from palaestrai.agent import SensorInformation, ActuatorInformation


@dataclass
class AgentSetupRequest:
    """Initializes the setup of an :class:`Agent`.

    * Sender: :class:`SimulationController`
    * Receiver: :class:`AgentConductor`

    :param sender_simulation_controller: ID of the sending
        :class:`SimulationController`
    :param receiver_agent_conductor: ID of the receiving
        :class:`AgentConductor`
    :param experiment_run_id: ID of the experiment run the agent participates
        in
    :param agent_id: ID of the agent we're setting up (e.g., a :class:`Muscle`)
    :param sensors: List of :class:`SensorInformation` objects for the
        sensors available to the agent
    :param actuators: List of :class:`ActuatorInformation` objects for the
        actuators available to the agent
    :param agent_name: Name of the :class:`Agent`, if any
    """

    sender_simulation_controller: str
    receiver_agent_conductor: str
    experiment_run_id: str
    agent_id: str
    sensors: List[SensorInformation]
    actuators: List[ActuatorInformation]
    agent_name: str

    @property
    def sender(self):
        return self.sender_simulation_controller

    @property
    def receiver(self):
        return self.receiver_agent_conductor



from typing import TYPE_CHECKING

from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from palaestrai.experiment import RunGovernor


class RGSStoppingTransceiving(RunGovernorState):
    """Represents the STOPPING_TRANSCEIVING state.

    Possible next states are
    * :class:`.RGSStoppingRun in the normal case.

    """

    def __init__(self, rgc: "RunGovernor"):
        super().__init__(rgc, "STOPPING_TRANSCEIVING")

    async def run(self):
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) about to shut down worker.",
            id(self.rgc),
            self.rgc.uid,
        )
        try:
            msg = self.rgc.next_reply.pop()
        except IndexError:
            msg = None

        _ = await self.rgc.major_domo_worker.transceive(msg, skip_recv=True)

    def next_state(self):
        from . import RGSStoppingRun

        self.rgc.state = RGSStoppingRun(self.rgc)



from .space import Space
from .box import Box
from .discrete import Discrete
from .multi_discrete import MultiDiscrete
from .multi_binary import MultiBinary
from .tuple import Tuple

from .utils import flatdim
from .utils import flatten
from .utils import unflatten

__all__ = [
    "Space",
    "Box",
    "Discrete",
    "MultiDiscrete",
    "MultiBinary",
    "Tuple",
    "flatdim",
    "flatten",
    "unflatten",
]



"""This module contains the abstract baseclass :class:`.Objective`,
from which all other objectives should be derived.

"""
from abc import ABC, abstractmethod
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from . import RewardInformation


class Objective(ABC):
    """The base class for all objectives.

    An objective defines the goal of an agent and changing the
    objective can, e.g., transform an attacker agent to a defender
    agent.

    The objective can, e.g., a wrapper for the reward of the
    environment and, in the easiest case, the sign of the reward
    is flipped (or not) to define attacker or defender. However, the
    objective can as well use a complete different
    formula.

    """

    def __init__(self, params: dict):
        self.params = params

    @abstractmethod
    def internal_reward(self, rewards: List["RewardInformation"]) -> float:
        """Calculate the reward of this objective."""
        raise NotImplementedError



import re
from abc import ABC, abstractmethod

import numpy as np

from palaestrai.util import seeding
from palaestrai.util.dynaloader import locate


class Space(ABC):
    """Base class for space definitions

    Derived classes allow a minimal mathematical representation of a space
    concept.
    """

    _SPACENAME_RE = re.compile(r"\A\s*?(\w+)\(([^;]+)\)\s*\Z")
    TYPES = [
        "Box",
        "Dict",
        "Discrete",
        "MultiBinary",
        "MultiDiscrete",
        "Tuple",
    ]

    def __init__(self, shape=None, dtype=None):
        self.shape = None if shape is None else tuple(shape)
        self.dtype = None if dtype is None else np.dtype(dtype)
        self.np_random = None
        self.seed()

    @abstractmethod
    def sample(self):
        """Uniformly randomly sample a random element of this space."""
        pass

    def seed(self, seed=None):
        """Seed the PRNG of this space."""
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    @abstractmethod
    def contains(self, x):
        """Return ``True`` if the value given is a valid member of the space.

        :param x: Any value

        :return: True iff ``x`` is a member of the space
        """

        pass

    def __contains__(self, x):
        return self.contains(x)

    @abstractmethod
    def to_string(self):
        """Returns the string representation of the space object.

        String representations of space objects must read like a Python
        constructor, such as ``Box(low=[1.0, 0.0], high=[42.0, 42.42])``. The
        name of the class is used to dynamically construct a new object from a
        string representation.

        :return: The object's string representation.
        """

        pass

    @classmethod
    def from_string(cls, s):
        """Dynamically constructs any descendant space object from a string.

        On the :py:class:`Space` class, this method acts as a factory to
        construct any known descrendant module from a string representation.
        It dynamically locates and instantiates any module that is listed in
        :py:data:`Space.TYPES`.

        All descendant types must also implement this class method as an
        alternative constructor (the known ones do). I.e., you can use
        ``palaestrai.types.Discrete.from_string("Discrete(2)")`` as well as
        ``palaestrai.types.Space.from_string("Discrete(2)")``.

        :param str s: The string representation of a space object.
        """

        if cls != Space:
            raise NotImplementedError()

        m = Space._SPACENAME_RE.match(s)
        classname = m[1] if m else None

        if not classname:
            raise RuntimeError(
                "String '%s' did not match pattern '%s'"
                % (s, Space._SPACENAME_RE)
            )
        if not classname in Space.TYPES:
            raise RuntimeError("Unknown type: palaestrai.types.%s" % classname)

        classpath = "palaestrai.types.%s" % classname
        cls = locate(classpath)
        return cls.from_string(s)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string()



"""This module contains the class :class:`ActuatorInformation` that
stores all information the agents need about an actuator.

"""

from __future__ import annotations

from typing import TYPE_CHECKING

from palaestrai.util.exception import OutOfActionSpaceError

if TYPE_CHECKING:
    from palaestrai.types import Space


class ActuatorInformation:
    """Stores information about a single actuator.

    The actuator information class is used to transfer actuator
    information. It can be called to set a new setpoint:

        a = Actuator(some_space)
        a(42)  # a.setpoint is now 42

    Parameters
    ----------
    action_space : :class:`~.Space`
        An instance of a palaestrai space that defines the type of
        the :attr:`setpoint`.
    setpoint : any, optional
        The set value for this actuator. The type is defined by the
        :attr:`action_space`. Can be skipped and set afterwards.
    actuator_id : int or str, optional
        A unique identifier for this actuator. The agents use this ID
        only to assign the setpoints to the correct actuator. The ID
        is not analyzed to gain domain knowledge.
    """

    def __init__(self, setpoint, action_space: Space, actuator_id=None):
        self._setpoint = setpoint
        self.actuator_id = actuator_id
        self.action_space = action_space

    @property
    def setpoint(self):
        return self._setpoint

    @setpoint.setter
    def setpoint(self, setpoint):
        if setpoint is not None and self.action_space is not None:
            try:
                contained = self.action_space.contains(setpoint)
                if not contained:
                    msg = (
                        f"Setpoint '{setpoint}' is not contained "
                        f"within space '{self.action_space}'."
                    )
            except ValueError as e:
                msg = e.with_traceback
                contained = False

            if not contained:
                raise OutOfActionSpaceError(msg)

            self._setpoint = setpoint

    def __call__(self, setpoint):
        self.setpoint = setpoint

    def __repr__(self):
        return (
            "ActuatorInformation("
            "setpoint=%s, action_space=%s, actuator_id=%s)"
            % (self.setpoint, repr(self.action_space), self.actuator_id)
        )

    @property
    def id(self):
        return self.actuator_id

    @id.setter
    def id(self, value):
        self.actuator_id = value



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class NextPhaseRequest:
    """A request that is send from a :class:`RunGovernor` to itself.

    Actually, this request is not sent at all. It is directly stored
    in the run governor's list of last requests (sending it via client
    results in an error because of time-related dependencies in the
    order of messages sent by the run governor).

    It is processed in the state STARTING_SETUP and initiates a new
    phase of an experiment run.

    Parameters
    ----------
    sender_run_governor_id: str
        ID of the sending run governor
    receiver_run_governor_id: str
        ID of the receiving run governer (it's the same as the sender).
    next_phase: int
        Index of the next phase to execute.
    """

    sender_run_governor_id: str
    receiver_run_governor_id: str
    next_phase: int

    @property
    def sender(self):
        """ID of the sending run governor."""
        return self.sender_run_governor_id

    @property
    def receiver(self):
        """ID of the receiving run governor."""
        return self.receiver_run_governor_id



import re
import numpy as np
from .space import Space


class MultiBinary(Space):
    """A special form of MultiDiscrete for multiple binary values"""

    _RE = re.compile(r"\AMultiBinary\(\s*(?P<inner>\d+)\s*\)\Z")

    def __init__(self, n):
        self.n = n
        super(MultiBinary, self).__init__((self.n,), np.int8)

    def sample(self):
        return self.np_random.randint(
            low=0, high=2, size=self.n, dtype=self.dtype
        )

    def contains(self, x):
        if isinstance(x, list):
            x = np.array(x)  # Promote list to array for contains check
        return (
            ((x == 0) | (x == 1)).all()
            and len(x) == self.n
            and (x.dtype == int)
        )

    def to_string(self):
        return "MultiBinary(%s)" % self.n

    @classmethod
    def from_string(cls, s):
        match = MultiBinary._RE.match(s)
        if not match or not match["inner"]:
            raise RuntimeError(
                "String '%s' did not match '%s'" % (s, MultiBinary._RE)
            )
        return MultiBinary(int(match["inner"]))

    def __repr__(self):
        return self.to_string()

    def __eq__(self, other):
        return isinstance(other, MultiBinary) and self.n == other.n



"""Add experiment name

Revision ID: 728eceef1424
Revises: 158c30d6a4bf
Create Date: 2020-11-26 20:31:17.452703

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "728eceef1424"
down_revision = "158c30d6a4bf"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "experiments",
        sa.Column("name", sa.VARCHAR(255), unique=True, index=True),
    )


def downgrade():
    op.drop_column("experiments", "name")



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class EnvironmentResetNotificationRequest:
    """Notifies an agent about an environment reset.

    Parameters
    ----------
    sender_simulation_controller_id: str
        ID of the sending :class:`.SimulationController`.
    receiver_agent_id: str
        ID of the receiving :class:`.Muscle`.

    """

    sender_simulation_controller_id: str
    receiver_agent_id: str

    @property
    def sender(self):
        return self.sender_simulation_controller_id

    @property
    def receiver(self):
        return self.receiver_agent_id



import unittest
from unittest.mock import AsyncMock, call, patch
from uuid import uuid4

from palaestrai.core.protocol import (
    EnvironmentSetupRequest,
    EnvironmentSetupResponse,
    ShutdownRequest,
    ShutdownResponse,
)
from palaestrai.environment.environment_conductor import EnvironmentConductor


class TestEnvironmentConductor(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.mockaio = patch(
            "palaestrai.environment.environment_conductor.aiomultiprocess."
            "Process",
            spec=True,
        ).start()

        self.addCleanup(patch.stopall)

        self.env_cond = EnvironmentConductor(
            {
                "name": (
                    "palaestrai.environment.dummy_environment:"
                    "DummyEnvironment"
                ),
                "params": dict(),
            },
            {
                "name": "palaestrai.environment.dummy_reward:DummyReward",
                "params": dict(),
            },
            "test_conn",
            123,
            uuid4(),
        )

    def test_one_process_per_environment(self):
        self.env_cond.init_env(uuid4())
        self.env_cond.init_env(uuid4())
        self.assertEqual(self.mockaio.call_count, 2)
        self.assertEqual(
            len(self.env_cond._tasks), 2, "not correctly executed"
        )

    async def test_run_shutdown(self):
        msg_shutdown = ShutdownRequest(42)
        self.env_cond.worker.transceive = AsyncMock(return_value=msg_shutdown)

        await self.env_cond.run()
        self.env_cond.worker.transceive.assert_awaited()

    async def test_setup_conductor(self):
        msg_setup = EnvironmentSetupRequest(
            environment_id="0815",
            experiment_run_id="run away",
            receiver_environment_conductor_id="the boss",
            sender_simulation_controller_id="the servant",
        )

        msg_setup_response = EnvironmentSetupResponse(
            sender_environment_conductor=self.env_cond.uid,
            receiver_simulation_controller="the servant",
            environment_id="0815",
            experiment_run_id="run away",
            environment_type=self.env_cond.env_cfg["name"],
            environment_parameters=self.env_cond.env_cfg["params"],
        )

        msg_shutdown = ShutdownRequest(42)
        msg_shutdown_response = ShutdownResponse(42)

        calls = (
            call(None),
            call(msg_setup_response),
            call(msg_setup_response),
            call(msg_shutdown_response, skip_recv=True),
        )

        self.env_cond.worker.transceive = AsyncMock(
            side_effect=[msg_setup, msg_setup, msg_shutdown, None]
        )

        await self.env_cond.run()
        self.assertEqual(len(self.env_cond._tasks), 2)
        self.env_cond._worker.transceive.assert_has_awaits(calls)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import MagicMock, patch

from palaestrai.agent import RewardInformation
from palaestrai.agent.dummy_brain import DummyBrain
from palaestrai.agent.dummy_objective import DummyObjective

from palaestrai.agent.state import State
from palaestrai.core.protocol import MuscleUpdateRequest, MuscleShutdownRequest
from palaestrai.core.serialisation import serialize
from palaestrai.types import Discrete


class TestBrain(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.brain_params = {}
        self.muscle_update_req = MuscleUpdateRequest(
            sender_muscle_id="0",
            receiver_brain_id="1",
            experiment_run_id="2",
            agent_id="3",
            sensor_readings=list(),
            last_actions=list(),
            reward=[RewardInformation(0.0, Discrete(1), "Test")],
            is_terminal=False,
            shutdown=False,
        )
        self.muscle_shutdown_req = MuscleShutdownRequest(
            sender_muscle_id="0",
            experiment_run_id="2",
            agent_id="3",
        )
        self.brain = DummyBrain(
            "test-con",
            list(),
            list(),
            DummyObjective({}),
            "./test123/phase_test/testAgent",
            123,
            layers=2,
        )

    def test_process(self):
        self.brain._state = State.RUNNING
        self.brain._connect = MagicMock()
        self.brain._router_socket = MagicMock()
        self.brain._router_socket.recv_multipart = MagicMock()
        update_msg = serialize(self.muscle_update_req)
        shutdown_msg = serialize(self.muscle_shutdown_req)
        self.brain._router_socket.recv_multipart.side_effect = [
            ["0", update_msg],
            ["0", shutdown_msg],
        ]
        self.brain._receive_updates()

        self.assertEqual(self.brain.worker_updates.qsize(), 2)
        self.assertFalse(self.brain._receive_updates())
        msg = self.brain.worker_updates.get()
        self.assertIsInstance(msg[0], MuscleUpdateRequest)
        self.assertEqual(msg[1], "0")

        # TODO: Change to MuscleShutdownRequest as soon as it is merged
        msg = self.brain.worker_updates.get()
        self.assertIsInstance(msg[0], MuscleShutdownRequest)
        self.assertEqual(msg[1], "0")

    @patch("palaestrai.agent.brain.Thread")
    async def test_run(self, mockthread):
        self.brain._state = State.RUNNING
        self.brain._send = MagicMock()
        self.brain.thinking = MagicMock(return_value=0)
        self.brain.worker_updates.put([self.muscle_update_req, "0"])
        self.brain.worker_updates.put([self.muscle_shutdown_req, "0"])

        await self.brain.run()

        mockthread.assert_called_once()
        self.brain.thinking.assert_called_once()
        self.assertEqual(self.brain._send.call_count, 2)
        self.assertNotEqual(self.brain.state, State.RUNNING)


if __name__ == "__main__":
    unittest.main()



import asyncio
from typing import TYPE_CHECKING

from palaestrai.core.protocol import (
    ExperimentRunShutdownRequest,
    ExperimentRunShutdownResponse,
    NextPhaseRequest,
    NextPhaseResponse,
    SimulationStopRequest,
    SimulationStopResponse,
)

from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from palaestrai.experiment import RunGovernor


class RGSStoppingSimulation(RunGovernorState):
    """Represents the STOPPING_SIMULATION state.

    This state is entered after the EXPERIMENT_RUNNING state if the run
    governor receives an experiment shutdown request. In this state, a
    reply is prepared and passed to the next state.

    Possible next states are
    * :class:`.RGSStoppingTransceiving in the normal case.
    (* ABORTING (error case), not handled yet)

    Parameters
    ----------
    ctx: :class:`.RunGovernor`
        The run governor instance that provides the context for this
        state.
    request: :class:`.ExperimentShutdownRequest`
        The experiment shutdown request that leads to this state.

    """

    def __init__(self, rgc: "RunGovernor"):
        super().__init__(rgc, "STOPPING_SIMULATION")

    async def run(self):
        if len(self.rgc.errors) > 0:
            timeout = 5
        else:
            timeout = None
        try:
            sc_terminated = await asyncio.wait_for(
                self._terminate_sim_controllers(), timeout
            )
            self._prepare_reply(sc_terminated)
        except asyncio.exceptions.TimeoutError:
            LOG.warning(
                "RunGovernor(id=0x%x, uid=%s) "
                "encountered a timeout while waiting for the "
                "remaining processes to shut down orderly; "
                "killing them now.",
                id(self.rgc),
                self.rgc.uid,
            )

    def next_state(self, *args, **kwargs):
        from . import RGSStoppingRun, RGSStoppingTransceiving

        if self.rgc.experiment_run.has_next_phase(self.rgc.current_phase):
            self.rgc.state = RGSStoppingRun(self.rgc)
        else:
            self.rgc.state = RGSStoppingTransceiving(self.rgc)

    async def _terminate_sim_controllers(self):
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) shutting down "
            "Experiment(id=%s, experiment_run_id=%s).",
            id(self.rgc),
            self.rgc.uid,
            self.rgc.experiment_run.user_id,
            self.rgc.experiment_run_id,
        )

        sc_terminated = []
        for sc_uid in self.rgc.sim_controllers:
            response = await self.rgc.major_domo_client.send(
                sc_uid, SimulationStopRequest()
            )
            if response is None or not isinstance(
                response, SimulationStopResponse
            ):
                self.rgc.sim_controllers[sc_uid].terminate()
                sc_terminated += [sc_uid]
                continue
            LOG.debug(
                "RunGovernor(id=0x%x, uid=%s) received "
                "SimulationStopResponse(experiment_run_id=%s); "
                "list of running experiments contains: %s.",
                id(self.rgc),
                self.rgc.uid,
                response.experiment_run_id,
                self.rgc.sim_controllers.keys(),
            )
        return sc_terminated

    def _prepare_reply(self, sc_terminated):
        try:
            request = self.rgc.last_request[0]
        except IndexError:
            request = None

        if isinstance(request, ExperimentRunShutdownRequest):
            self.rgc.next_reply.append(
                ExperimentRunShutdownResponse(
                    sender_run_governor_id=self.rgc.uid,
                    receiver_executor_id=request.sender,
                    experiment_run_id=self.rgc.experiment_run_id,
                    successful=len(sc_terminated) == 0,
                    error=(
                        None
                        if len(sc_terminated) == 0
                        else "Had to terminate: %s" % sc_terminated
                    ),
                )
            )
            self.rgc.last_request.pop(0)
        else:
            has_next_phase = isinstance(request, NextPhaseRequest)
            self.rgc.next_reply.append(
                NextPhaseResponse(
                    sender_run_governor_id=self.rgc.uid,
                    receiver_run_governor_id=self.rgc.uid,
                    has_next_phase=has_next_phase,
                )
            )



import unittest
from unittest.mock import MagicMock, PropertyMock

from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import (
    RGSHandlingDeadChildren,
    RGSStoppingSimulation,
    RGSStoppingTransceiving,
)
from palaestrai.util.exception import PrematureTaskDeathError


class TestHandlingDeadChildren(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)
        self.rgs = RGSHandlingDeadChildren(self.rungov)
        self.rungov.state = self.rgs

    async def test_run(self):
        sim = MagicMock()
        sim.is_alive = MagicMock(return_value=True)
        type(sim).exit_code = PropertyMock(return_value=0)
        self.rungov.sim_controllers["sim-1"] = sim

        await self.rgs.run()

        sim.is_alive.assert_called_once()

    async def test_run_alive_false(self):
        sim = MagicMock()
        sim.is_alive = MagicMock(return_value=False)
        type(sim).exit_code = PropertyMock(return_value=0)
        self.rungov.sim_controllers["sim-1"] = sim

        await self.rgs.run()

        self.assertEqual(2, sim.is_alive.call_count)
        self.assertEqual(0, len(self.rungov.sim_controllers))

    async def test_run_non_zero_exitcode(self):
        sim = MagicMock()
        sim.is_alive = MagicMock(return_value=False)
        type(sim).exit_code = PropertyMock(return_value=1)
        self.rungov.sim_controllers["sim-1"] = sim

        await self.rgs.run()

        self.assertEqual(2, sim.is_alive.call_count)
        self.assertEqual(0, len(self.rungov.sim_controllers))
        self.assertIsInstance(
            self.rungov.errors[0][0], PrematureTaskDeathError
        )

    def test_next_state(self):

        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSStoppingSimulation)

    def test_next_state_with_errors(self):
        self.rgs.add_error(ValueError())

        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSStoppingTransceiving)


if __name__ == "__main__":
    unittest.main()



from unittest import TestCase
import numpy as np
from palaestrai.types import Box
from palaestrai.types import Discrete
from palaestrai.types import MultiDiscrete
from palaestrai.types import MultiBinary
from palaestrai.types import Tuple
from palaestrai.types import utils


class UtilsTest(TestCase):
    def setUp(self) -> None:
        self.box_float = Box(
            low=-1.0, high=2.0, shape=(3, 4), dtype=np.float32
        )
        self.discrete_sample = Discrete(3)
        self.multi_binary_sample = MultiBinary(2)
        self.multi_discrete_sample = MultiDiscrete([5, 4, 2])
        self.tuple_discrete_sample = Tuple(Discrete(2), Discrete(3))
        self.tuple_boxes_sample = Tuple(
            Box(low=-1.0, high=2.0, shape=(3, 4), dtype=np.float32),
            Box(low=-2.0, high=3.0, shape=(2, 3), dtype=np.float32),
        )

    def test_flatdim(self):
        self.assertEqual(utils.flatdim(self.box_float), 12)
        self.assertEqual(utils.flatdim(self.discrete_sample), 3)
        self.assertEqual(utils.flatdim(self.tuple_discrete_sample), 5)
        self.assertEqual(utils.flatdim(self.tuple_boxes_sample), 18)
        self.assertEqual(utils.flatdim(self.multi_binary_sample), 2)
        self.assertEqual(utils.flatdim(self.multi_discrete_sample), 3)
        with self.assertRaises(NotImplementedError):
            utils.flatdim("any other type")

    def test_flatten_Box(self):
        x = [
            [1.9454306, 0.25988454, -0.41315162, 1.3343621],
            [1.7871696, -0.6225805, 0.88759154, 0.5394544],
            [1.2164323, 0.21180445, 1.0201195, -0.4387453],
        ]
        x_expected = np.array(
            [
                1.9454306,
                0.25988454,
                -0.41315162,
                1.3343621,
                1.7871696,
                -0.6225805,
                0.88759154,
                0.5394544,
                1.2164323,
                0.21180445,
                1.0201195,
                -0.4387453,
            ],
            dtype=np.float32,
        )

        self.assertEqual(
            utils.flatten(self.box_float, x).shape, x_expected.shape
        )
        for pair in zip(x_expected, utils.flatten(self.box_float, x)):
            self.assertEqual(pair[0], pair[1])

    def test_flatten_discrete(self):
        x = 2
        x_expected = [0.0, 0.0, 1.0]

        for pair in zip(x_expected, utils.flatten(self.discrete_sample, x)):
            self.assertEqual(pair[0], pair[1])

    def test_flatten_Tuple(self):
        # for tuple of discrete:
        t = (1, 2)
        t_expected = [0.0, 1.0, 0.0, 0.0, 1.0]
        for pair in zip(
            t_expected, utils.flatten(self.tuple_discrete_sample, t)
        ):
            self.assertEqual(pair[0], pair[1])
        # for tuple of boxes:
        tboxes = (
            [
                [1.9454306, 0.25988454, -0.41315162, 1.3343621],
                [1.7871696, -0.6225805, 0.88759154, 0.5394544],
                [1.2164323, 0.21180445, 1.0201195, -0.4387453],
            ],
            [
                [1.9454306, 0.25988454, -0.41315162],
                [1.7871696, -0.6225805, 0.88759154],
            ],
        )
        tboxes_expected = np.array(
            [
                1.9454306,
                0.25988454,
                -0.41315162,
                1.3343621,
                1.7871696,
                -0.6225805,
                0.88759154,
                0.5394544,
                1.2164323,
                0.21180445,
                1.0201195,
                -0.4387453,
                1.9454306,
                0.25988454,
                -0.41315162,
                1.7871696,
                -0.6225805,
                0.88759154,
            ],
            dtype=np.float32,
        )

        self.assertEqual(
            utils.flatten(self.tuple_boxes_sample, tboxes).shape,
            tboxes_expected.shape,
        )
        for pair in zip(
            tboxes_expected, utils.flatten(self.tuple_boxes_sample, tboxes)
        ):
            self.assertEqual(pair[0], pair[1])

    def test_flatten_multibinary(self):
        x = np.array([0, 1], dtype=int)
        x_expected = np.array([0, 1], dtype=int)
        for pair in zip(
            x_expected, utils.flatten(self.multi_binary_sample, x)
        ):
            self.assertEqual(pair[0], pair[1])

    def test_flatten_multidiscrete(self):
        x = np.array([1, 1, 0], dtype=int)
        x_expected = np.array(
            [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0],
            dtype=np.float32,
        )
        for pair in zip(
            x_expected, utils.flatten(self.multi_discrete_sample, x)
        ):
            self.assertEqual(pair[0], pair[1])

    def test_flatten_othertypes(self):
        with self.assertRaises(NotImplementedError):
            utils.flatten("other space", "any other type")

    def test_unflatten_Box(self):
        x = np.array(
            [
                1.9454306,
                0.25988454,
                -0.41315162,
                1.3343621,
                1.7871696,
                -0.6225805,
                0.88759154,
                0.5394544,
                1.2164323,
                0.21180445,
                1.0201195,
                -0.4387453,
            ],
            dtype=np.float32,
        )

        x_expected = np.array(
            [
                [1.9454306, 0.25988454, -0.41315162, 1.3343621],
                [1.7871696, -0.6225805, 0.88759154, 0.5394544],
                [1.2164323, 0.21180445, 1.0201195, -0.4387453],
            ],
            dtype=np.float32,
        )

        self.assertEqual(
            utils.unflatten(self.box_float, x).shape, x_expected.shape
        )
        for pair in zip(utils.unflatten(self.box_float, x), x_expected):
            self.assertEqual(np.array_equal(pair[0], pair[1]), True)

    def test_unflatten_Discrete(self):

        x = [0.0, 0.0, 1.0]
        x_expected = 2
        self.assertEqual(x_expected, utils.unflatten(self.discrete_sample, x))

    def test_unflatten_Tuple(self):
        # for tuple of discrete:
        t = [0.0, 1.0, 0.0, 0.0, 1.0]
        t_expected = (1, 2)
        for pair in zip(
            t_expected, utils.unflatten(self.tuple_discrete_sample, t)
        ):
            self.assertEqual(pair[0], pair[1])
        # for tuple of boxes:
        tboxes = np.array(
            [
                1.9454306,
                0.25988454,
                -0.41315162,
                1.3343621,
                1.7871696,
                -0.6225805,
                0.88759154,
                0.5394544,
                1.2164323,
                0.21180445,
                1.0201195,
                -0.4387453,
                1.9454306,
                0.25988454,
                -0.41315162,
                1.7871696,
                -0.6225805,
                0.88759154,
            ],
            dtype=np.float32,
        )
        tboxes_expected = (
            np.array(
                [
                    [1.9454306, 0.25988454, -0.41315162, 1.3343621],
                    [1.7871696, -0.6225805, 0.88759154, 0.5394544],
                    [1.2164323, 0.21180445, 1.0201195, -0.4387453],
                ],
                dtype=np.float32,
            ),
            np.array(
                [
                    [1.9454306, 0.25988454, -0.41315162],
                    [1.7871696, -0.6225805, 0.88759154],
                ],
                dtype=np.float32,
            ),
        )

        # both shape and values are checked in np.array_equal
        for pair in zip(
            tboxes_expected, utils.unflatten(self.tuple_boxes_sample, tboxes)
        ):
            self.assertEqual(np.array_equal(pair[0], pair[1]), True)

    def test_unflatten_multibinary(self):
        x = np.array([0, 1], dtype=int)
        x_expected = np.array([0, 1], dtype=int)  # trivial ?!
        for pair in zip(
            x_expected, utils.unflatten(self.multi_binary_sample, x)
        ):
            self.assertEqual(pair[0], pair[1])

    def test_unflatten_multidiscrete(self):
        x = np.array(
            [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0],
            dtype=np.float32,
        )
        x_expected = np.array([4, 1, 0], dtype=int)
        for pair in zip(
            x_expected, utils.unflatten(self.multi_discrete_sample, x)
        ):
            self.assertEqual(pair[0], pair[1])

    def test_unflatten_othertypes(self):
        with self.assertRaises(NotImplementedError):
            utils.unflatten("other space", "any other type")



"""This module contains the class :class:`AgentConductor` that
stores all information the agents need about an actuator.

"""
from __future__ import annotations

import asyncio
import logging
import signal
import uuid
from typing import Union, List
from uuid import uuid4

import aiomultiprocess
import setproctitle

from palaestrai.core import MajorDomoWorker, RuntimeConfig
from palaestrai.core.protocol import (
    AgentSetupRequest,
    AgentSetupResponse,
    ShutdownRequest,
    ShutdownResponse,
)
from palaestrai.util import spawn_wrapper
from palaestrai.util.dynaloader import load_with_params
from palaestrai.util.exception import TasksNotFinishedError
from .brain import Brain
from .muscle import Muscle

LOG = logging.getLogger("palaestrai.agent.conductor")


async def _run_brain(agent_conductor_uid: str, brain: Brain):
    """Executes the Brain main loop, handling signals and errors

    This method wraps :py:func:`Brain.run`. It takes care of proper
    installment of signal handlers, setting of the proctitle, and most
    importantly, error handling. This method should be wrapped in the
    :py:func:`palaestrai.util.spawn_wrapper` function, which, in turn, is the
    target of an :py:func:`aiomultiprocess.Process.run` call.

    Parameters
    ----------
    agent_conductor_uid : str
        UID of the governing ::`~AgentConductor`
    brain : Brain
        The ::`~Brain` instance that should be run.

    Returns
    -------
    Any
        Whatever the ::`~Brain.run` method returns.
    """
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGABRT, signal.SIG_DFL)
    signal.signal(signal.SIGTERM, signal.SIG_DFL)
    setproctitle.setproctitle(
        "palaestrAI[Brain-%s]" % agent_conductor_uid[-6:]
    )
    try:
        return await brain.run()
    except Exception as e:
        LOG.critical(
            "Brain(id=0x%x, agent_conductor_uid=%s) died from a fatal wound "
            "to the head: %s",
            id(brain),
            agent_conductor_uid,
            e,
        )
        raise


async def _run_muscle(muscle: Muscle):
    """Executes the ::`~Muscle` main loop, handling signals and errors

    This method wraps :py:func:`Muscle.run`. It takes care of proper
    installment of signal handlers, setting of the proctitle, and most
    importantly, error handling. This method should be wrapped in the
    :py:func:`palaestrai.util.spawn_wrapper` function, which, in turn, is the
    target of an :py:func:`aiomultiprocess.Process.run` call.

    Parameters
    ----------
    muscle : Muscle
        The  ::`~Muscle` instance that runs

    Returns
    -------
    Any
        Whatever ::`~Muscle.run` returns
    """
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGABRT, signal.SIG_DFL)
    signal.signal(signal.SIGTERM, signal.SIG_DFL)
    setproctitle.setproctitle("palaestrAI[Muscle-%s]" % muscle.uid[-6:])
    try:
        return await muscle.run()
    except Exception as e:
        LOG.critical(
            "Muscle(id=0x%x, uid=%s) suffers from dystrophy: %s",
            id(muscle),
            muscle.uid,
            e,
        )
        raise


class AgentConductor:
    """This creates a new agent conductor (AC).

    The AC receives an agent config, which contains all information for
    the brain and the muscle. Additional information, like the current
    run ID, are part of the AgentSetupRequest.

    Parameters
    ----------
    broker_uri: str
        Connection information, should be in the format
        protocol://ip-address:port, e.g., "tcp://127.0.0.1:45310"
    agent_config: dict
        A *dict* containing information, how to instantiate brain and
        muscle.
    seed: int
        The random seed for this agent conductor.
    phase_name: str
        Name of the phase the experiment is currently in.
    experiment_run_id: str
        The ID of the overall experiment of which this brain is part of.
    uid: str
        The uid for this agent conductor object.

    Attributes
    ----------
    agents: List[str]
        A list containing the uids of all the muscles.
    brain: :class:`.Brain`:
        A reference to the brain instance of this agent conductor.
    brain_process: :class:`aiomultiprocess.Process`
        A reference to the brain's process.
    objective: :class:`.Objective`
        A reference to the objective instance of this agent conductor.
    tasks: List[aiomultiprocess.Process]
        A list with references to the muscle's tasks.
    """

    def __init__(
        self,
        broker_uri: str,
        agent_config: dict,
        seed: int,
        phase_name: str,
        experiment_run_id: str = None,
        uid=None,
    ):
        self._worker = None
        self._broker_uri = broker_uri
        self._uid = str(uid) if uid else "AgentConductor-%s" % uuid4()

        self.tasks: List[aiomultiprocess.Process] = []
        self.seed = seed
        self.conf = agent_config
        self.brain: Union[Brain, None] = None
        self.objective = load_with_params(
            self.conf["objective"]["name"], self.conf["objective"]["params"]
        )
        self.agents: List[uuid.UUID] = []

        self.ac_socket = None
        self.brain_process = None
        self.experiment_run_id = experiment_run_id
        self.phase_name = phase_name

    def _handle_sigintterm(self, signum, frame):
        LOG.info(
            "AgentConductor(id=0x%x, uid=%s) "
            "interrupted by signal %s in frame %s",
            id(self),
            self.uid,
            signum,
            frame,
        )
        raise SystemExit()

    @property
    def uid(self):
        """Unique, opaque ID of the agent conductor object"""
        return self._uid

    @property
    def worker(self):
        """Getter for the :py:class:`MajorDomoWorker` object

        This method returns (possibly lazily creating) the current
        :py:class:`MajorDomoWorker` object. It creates this worker on demand.
        It is not safe to call this method between forks, as forks copy
        the context information for the worker which is process-depentent.

        :rtype: MajorDomoWorker
        """
        if self._worker is None:
            self._worker = MajorDomoWorker(self._broker_uri, self.uid)
        return self._worker

    def _init_brain(self, sensors, actuators):
        """Initialize the brain process.

        Each agent, which is represented by an individual conductor,
        has one brain process. This function initializes the brain
        process.

        Parameters
        ----------
        sensors : list[sensor]
            List of available sensors.
        actuators : list[actuator]
            List of available actuators.
        """
        params = self.conf["brain"]["params"]
        params["sensors"] = sensors
        params["actuators"] = actuators
        params["objective"] = self.objective
        params["seed"] = self.seed
        params["store_path"] = self.__construct_store_path(self.conf)
        self.brain: Brain = load_with_params(
            self.conf["brain"]["name"], params
        )
        if "load" in self.conf.keys():
            load_path = self._construct_load_path(self.conf)
            if load_path is not None:
                self.brain.load_model(load_path)

        try:
            self.brain_process = aiomultiprocess.Process(
                name=f"Brain-{self.uid}",
                target=spawn_wrapper,
                args=(
                    f"Brain-{self.uid}",
                    RuntimeConfig().to_dict(),
                    _run_brain,
                    [self.uid, self.brain],
                ),
            )
            self.brain_process.start()
            LOG.debug(
                "AgentConductor(id=0x%x, uid=%s) started Brain(id=0x%x)",
                id(self),
                self.uid,
                id(self.brain),
            )
        except Exception as e:
            LOG.critical(
                "AgentConductor(id=0x%x, uid=%s) "
                "encountered a fatal error while executing "
                "Brain(id=0x%x): %s",
                id(self),
                self.uid,
                id(self.brain),
                e,
            )
            raise

    def _init_muscle(self, uid):
        """Function to initialize a new muscle

        Each agent consists of one brain and at least one muscle
        but is not limited to one muscle. There can be multiple
        muscles and muscles can be restarted.

        Parameters
        ----------
        uid : uuid4
            Unique identifier to identify the muscle
        """
        params = self.conf["muscle"]["params"]
        params["uid"] = uid
        params["brain_id"] = id(self.brain)
        muscle = load_with_params(self.conf["muscle"]["name"], params)
        muscle.setup()
        self.agents.append(uid)
        try:
            agent_process = aiomultiprocess.Process(
                name=f"Muscle-{uid}",
                target=spawn_wrapper,
                args=(
                    f"Muscle-{uid}",
                    RuntimeConfig().to_dict(),
                    _run_muscle,
                    [muscle],
                ),
            )
            agent_process.start()
            LOG.debug(
                "AgentConductor(id=0x%x, uid=%s) "
                "started Muscle(id=0x%x, uid=%s).",
                id(self),
                self.uid,
                id(muscle),
                muscle.uid,
            )
            self.tasks.append(agent_process)
        except Exception as e:
            LOG.critical(
                "AgentConductor(id=0x%x, uid=%s) "
                "encountered a fatal error while executing "
                "Muscle(id=0x%x, uid=%s): %s",
                id(self),
                self.uid,
                id(muscle),
                muscle.uid,
                e,
            )
            raise e

    async def run(self):
        """Monitors agents and facilitates information interchange

        This method is the main loop for the :py:class:`AgentConductor`. It
        monitors the :py:class:`Brain` object and :py:class:`Muscle` instances
        of the agent (i.e., the processes) and transceives/routes messages.

        """
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGABRT, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        setproctitle.setproctitle(
            "palaestrAI[AgentConductor-%s]" % self._uid[-6:]
        )

        signal.signal(signal.SIGINT, self._handle_sigintterm)
        signal.signal(signal.SIGTERM, self._handle_sigintterm)
        LOG.info(
            "AgentConductor(id=0x%x, uid=%s) commencing run: "
            "Today's solutions to tomorrow's problems",
            id(self),
            self.uid,
        )
        proceed = True
        request = None
        reply = None
        while proceed:
            try:
                request = await self._housekeeping(reply)
            except TasksNotFinishedError:
                continue
            except SystemExit:
                break

            if isinstance(request, AgentSetupRequest):
                reply = self._handle_agent_setup(request)

            if isinstance(request, ShutdownRequest):
                await self._handle_shutdown(request)
                break

        LOG.debug(
            "AgentConductor(id=0x%x, uid=%s) sending "
            "ShutdownResponse(experiment_run_id=%s)",
            id(self),
            self.uid,
            request.run_id,
        )
        reply = ShutdownResponse(request.run_id)
        try:
            await self.worker.transceive(reply, skip_recv=True)
        except SystemExit:
            pass  # If they really want to, we can skip that, too.
        LOG.info(
            "AgentConductor(id=0x%x, uid=%s) completed shutdown: "
            "ICH, AC, BIN NUN TOD, ADJÖ [sic].",
            id(self),
            self.uid,
        )

    async def _housekeeping(self, reply):
        """Keep the household clean and lively.

        In this method, replies are send and requests are received.
        Furthermore, the AC sees over his child tasks (muscles and
        brain).

        Parameters
        ----------
        reply:
            The next reply to send

        Returns
        -------
        request
            The request received during transceiving.

        """
        LOG.debug(
            "AgentConductor(id=0x%x, uid=%s) starts housekeeping. "
            "Everything needs to be in proper order.",
            id(self),
            self.uid,
        )
        try:
            transceive_task = asyncio.create_task(
                self.worker.transceive(reply)
            )
            muscle_tasks = [asyncio.create_task(p.join()) for p in self.tasks]
            brain_tasks = (
                [asyncio.create_task(self.brain_process.join())]
                if self.brain_process
                else []
            )
            tasks_done, tasks_pending = await asyncio.wait(
                [transceive_task] + muscle_tasks + brain_tasks,
                return_when=asyncio.FIRST_COMPLETED,
            )
            if not tasks_done:
                # This shouldn't happen, but you never know.
                raise TasksNotFinishedError()

            terminated_workers = [
                p
                for p in self.tasks + [self.brain_process]
                if p is not None and not p.is_alive() and p.exitcode != 0
            ]
            if terminated_workers:
                # I don't think the other tasks should end like this?
                LOG.critical(
                    "AgentConductor(id=0x%x, uid=%s) "
                    "has suffered from prematurely dead tasks: %s",
                    id(self),
                    self.uid,
                    [p.name for p in terminated_workers],
                )
                signal.signal(signal.SIGINT, signal.SIG_DFL)
                signal.signal(signal.SIGABRT, signal.SIG_DFL)
                signal.signal(signal.SIGTERM, signal.SIG_DFL)
                raise RuntimeError(
                    "AgentConductor(id=0x%x, uid=%s) "
                    "has dead tasks at hands: %s"
                    % (
                        id(self),
                        self.uid,
                        [p.name for p in terminated_workers],
                    )
                )
            if transceive_task not in tasks_done:
                await transceive_task
            request = transceive_task.result()
        except SystemExit as e:
            LOG.warning(
                "AgentConductor(id=0x%x, uid=%s) "
                "stopping due to SIGINT/SIGTERM",
                id(self),
                self.uid,
            )
            raise e

        LOG.debug(
            "AgentConductor(id=0x%x, uid=%s) got a %s. "
            "Let's see how we handle this one.",
            id(self),
            self.uid,
            request,
        )

        return request

    def _handle_agent_setup(self, request):
        """Handle the agent setup request.

        One setup request will result in one new muscle created.
        The brain will be created if necessary.

        Parameters
        ----------
        request: :class:`.AgentSetupRequest`
            The agent setup request with information for the muscle to
            be created.

        Returns
        -------
        :class:`.AgentSetupResponse`
            The response for the simulation controller.

        """
        if request.receiver_agent_conductor == self.uid:
            if self.brain is None:
                self._init_brain(request.sensors, request.actuators)
            self._init_muscle(request.agent_id)

            return AgentSetupResponse(
                sender_agent_conductor=self.uid,
                receiver_simulation_controller=request.sender,
                experiment_run_id=request.experiment_run_id,
                agent_id=request.agent_id,
            )

    def _construct_load_path(self, conf: dict) -> Union[str, None]:
        """Construct the path specified by conf

        If required keys are not present an attempt is made
        to use a default value

        If the path could not be constructed a warning is logged

        :param conf: The config for the agent conductor
            The config needs the key 'load'
                load should be a dictionary that has at least 'phase_name' and may have
                       key      |                description            |        default
                'base'          : determines the base path to save to   : '.'
                'experiment_id' : the experiment_id to load from        : current experiment_id
                'agent_name'    : the name of the agent to load from    : same as the loading agent
            and 'name' which is the default fallback for agent_name
        :return: The path to load from on success, None otherwise
        """
        if "load" in conf:
            load_conf = conf["load"]
        else:
            return None

        path = ""

        if "base" in load_conf.keys():
            path += load_conf["base"]
        else:
            path += "."

        if "experiment_id" in load_conf.keys():
            path += f"/{load_conf['experiment_id']}"
        elif self.experiment_run_id is not None:
            path += f"/{self.experiment_run_id}"
        else:
            LOG.warning(
                "AgentConductor(id=0x%x, uid=%s) got a load_conf but 'experiment_id' is not provided. "
                "THE BRAIN WILL NOT LOAD!",
                id(self),
                self.uid,
            )
            return None

        if "phase_name" in load_conf.keys():
            path += f"/{load_conf['phase_name']}"
        else:
            LOG.warning(
                "AgentConductor(id=0x%x, uid=%s) got a load_conf but 'phase_name' is not provided. "
                "THE BRAIN WILL NOT LOAD!",
                id(self),
                self.uid,
            )
            return None

        if "agent_name" in load_conf.keys():
            path += f"/{load_conf['agent_name']}"
        else:
            path += f"/{conf['name']}"

        return path

    def __construct_store_path(self, conf: dict) -> str:
        """Construct the path specified by conf

        :param conf: The config for the agent conductor
            The config needs the key 'name' which is the name of
            the agent that stores
            It can have the key conf["brain"]["params"]["store_path"]
            to specify a base path to store to
        :return: The path to save to
        """
        path = ""

        if "store_path" in conf["brain"]["params"].keys():
            path += conf["brain"]["params"]["store_path"]
        else:
            path += "."

        path += f"/{self.experiment_run_id}/{self.phase_name}/{conf['name']}"

        return path

    async def _handle_shutdown(self, request):
        """Handle the shutdown request for this agent conductor.

        It is expected that all muscles and the brain of this
        agent conductor already received a shutdown request. Therefore,
        all this method does is to wait(join) for the processes.

        Parameters
        ----------
        request: :class:`.ShutdownRequest`
            The shutdown request
        """
        for task in self.tasks:
            await task.join()
        await self.brain_process.join()



import hashlib
from typing import Union, Optional, Tuple, cast, List

import os

from numpy.random import RandomState
import struct


def np_random(seed: Optional[int] = None) -> Tuple[RandomState, int]:
    if seed is not None and not (isinstance(seed, int) and 0 <= seed):
        raise RuntimeError(
            "Seed must be a non-negative integer "
            "or omitted, not {}".format(seed)
        )

    seed = create_seed(seed)

    rng: RandomState = RandomState()
    rng.seed(_int_list_from_bigint(hash_seed(seed)))
    return rng, seed


def hash_seed(seed: Optional[int] = None, max_bytes: int = 8) -> int:
    """Any given evaluation is likely to have many PRNG's active at
    once. (Most commonly, because the environment is running in
    multiple processes.) There's literature indicating that having
    linear correlations between seeds of multiple PRNG's can correlate
    the outputs:

    http://blogs.unity3d.com/2015/01/07/a-primer-on-repeatable-random-numbers/
    http://stackoverflow.com/questions/1554958/how-different-do-random-seeds-need-to-be
    http://dl.acm.org/citation.cfm?id=1276928

    Thus, for sanity we hash the seeds before using them. (This scheme
    is likely not crypto-strength, but it should be good enough to get
    rid of simple correlations.)

    Args:
        seed (Optional[int]): None seeds from an operating system specific randomness source.
        max_bytes: Maximum number of bytes to use in the hashed seed.
    """
    if seed is None:
        seed = create_seed(max_bytes=max_bytes)
    hsh = hashlib.sha512(str(seed).encode("utf8")).digest()
    return _bigint_from_bytes(hsh[:max_bytes])


def create_seed(a: Union[int, str, None] = None, max_bytes: int = 8) -> int:
    """Create a strong random seed. Otherwise, Python 2 would seed using
    the system time, which might be non-robust especially in the
    presence of concurrency.

    Args:
        a (Union[int, str, None]): None seeds from an operating system specific randomness source.
        max_bytes: Maximum number of bytes to use in the seed.
    """
    # Adapted from https://svn.python.org/projects/python/tags/r32/Lib/random.py
    if a is None:
        a = _bigint_from_bytes(os.urandom(max_bytes))
    elif isinstance(a, str):
        try:
            a = int_to_seed(int(a), max_bytes)
        except ValueError:
            a = cast(str, a)
            a_bytes = a.encode("utf8")
            a_bytes = a_bytes + hashlib.sha512(a_bytes).digest()
            a = _bigint_from_bytes(a_bytes[:max_bytes])
    elif isinstance(a, int):
        a = int_to_seed(a, max_bytes)
    else:
        raise RuntimeError("Invalid type for seed: %s (%s)" % (type(a), a))

    return a


def int_to_seed(a: int, max_bytes: int) -> int:
    """Turns integer into a seed.

    Parameters
    ----------
    a           Seed starting point
    max_bytes   Maximum number of bytes to use in the seed

    Returns
    -------
    a           Seed
    """
    a = a % 2 ** (8 * max_bytes)
    return a


def _bigint_from_bytes(bytes_: bytes) -> int:
    sizeof_int = 4
    padding = sizeof_int - len(bytes_) % sizeof_int
    bytes_ += b"\0" * padding
    int_count = int(len(bytes_) / sizeof_int)
    unpacked = struct.unpack("{}I".format(int_count), bytes_)
    accum = 0
    for i, val in enumerate(unpacked):
        accum += 2 ** (sizeof_int * 8 * i) * val
    return accum


def _int_list_from_bigint(bigint: int) -> List[int]:
    # Special case 0
    if bigint < 0:
        raise TypeError("Seed must be non-negative, not {}".format(bigint))
    elif bigint == 0:
        return [0]

    ints = []
    while bigint > 0:
        bigint, mod = divmod(bigint, 2 ** 32)
        ints.append(mod)
    return ints



import asyncio
import logging
import logging.config
import logging.handlers
import os
from pathlib import Path
from typing import Callable, Any, Union

from . import LOG
from ..core import RuntimeConfig


def _set_proctitle(process_name: str):
    try:
        import setproctitle

        setproctitle.setproctitle(f"palaestrAI[{process_name}")
    except ImportError:
        pass


def _restore_runtime_configuration(runtime_configuration_dict: dict):
    RuntimeConfig().reset()
    RuntimeConfig().load(runtime_configuration_dict)


def _reinitialize_logging():
    try:
        logging.config.dictConfig(RuntimeConfig().logging)
        logging.root.handlers.clear()
        logging.root.addHandler(
            logging.handlers.SocketHandler(
                "127.0.0.1", RuntimeConfig().logger_port
            )
        )
        logging.debug(
            "Reinitialized logging from RuntimeConfig(%s)", RuntimeConfig()
        )
    except (KeyError, ValueError) as e:
        logging.basicConfig(level=logging.INFO)
        logging.warning(
            "Could not load logging config (%s), continuing with defaults",
            e,
        )


async def spawn_wrapper(
    name: str,
    runtime_config: dict,
    callee: Callable,
    args: Union[list, None] = None,
    kwargs: Union[dict, None] = None,
) -> Any:
    """Wraps a target for fork/spawn and takes care of initialization.

    Whenever a new subprocess is created (regardless of whether spawn, fork,
    or forkserver is used), some caretaking needs to be done:

    * The runtime configuration needs to be transferred, and the
        ::`RuntimeConfig` properly reinitialized
    * Logging is reinitialized/rewired to send messages to the parent process
    * A proctitle is set

    Parameters
    ----------
    * name : str
        Name of the process; will lead to a proctitle in the form of
        ``palaestrai[%s]``
    * runtime_config : dict
        Runtime configuration dict, normally obtained from
        ::`RuntimeConfig.to_dict`
    * callee : Callable
        The target method
    * args : list, optional
        Positional arguments of ::`callee`.
    * kwargs : dict, optional
        Keyword arguments of ::`callee`

    Returns
    -------
    Any
        Whatever the target function returns.
    """
    if name:
        _set_proctitle(name)
    if not args:  # [] as default arg is mutable, workaround with None:
        args = []
    if not kwargs:  # {} as default arg is mutable, workaround with None:
        kwargs = {}
    _restore_runtime_configuration(runtime_config)
    _reinitialize_logging()

    profiler = None
    if RuntimeConfig().profile:
        import cProfile

        profiler = cProfile.Profile()
        profiler.enable()
    try:
        if asyncio.iscoroutinefunction(callee):
            ret = await callee(*args, **kwargs)
        else:
            ret = callee(*args, **kwargs)
    except Exception as e:
        LOG.critical("Running %s failed: %s", str(callee), e)
        raise e
    if profiler:
        profiler.disable()
        profiler.dump_stats(Path(os.curdir) / f"{name}.prof")
    return ret



"""
A simple, built-in logging server
"""

from __future__ import annotations

import asyncio
import logging
import pickle
import struct
from typing import List, Tuple

LOG = logging.getLogger(__name__)


class LogServer:
    """A simple, internal logging server that reinjects remote log messages

    Each submodule of palaestrAI that gets spawned lives in a separate
    process. As the ::`~spawn_wrapper` takes care of reinitializing the
    logger, it replaces all defined log handlers with a
    :py:class:`logging.SocketHandler`. This log server is ran by the
    ::`Executor` and receives all those messages. They are re-injected in the
    main process' logging system and treated according to the original
    logging configuration.
    """

    def __init__(self, listen_host: str, listen_port: int):
        """Constructs a new log server for a given address and port

        Parameters
        ----------
        listen_host : str
            The address the log server should bind to
        listen_port : int
            The port the log server should bind to
        """
        self._listen_host = listen_host
        self._listen_port = listen_port
        self._server = None
        self._clients: List[Tuple] = []

    def _add_client(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        LOG.debug(
            "LogServer(id=0x%x) registered a new client (%s, %s).",
            id(self),
            reader,
            writer,
        )
        self._clients.append(
            (reader, asyncio.create_task(LogServer._read_from_client(reader)))
        )

    @staticmethod
    async def _read_from_client(reader: asyncio.StreamReader):
        while True:
            chunk = await reader.read(4)
            msglen = struct.unpack(">L", chunk)[0]
            chunk = await reader.read(msglen)
            while len(chunk) < msglen:
                chunk = chunk + await reader.read(msglen - len(chunk))
            try:
                logobj = pickle.loads(chunk)
                if not logobj:
                    continue
            except:
                continue
            record = logging.makeLogRecord(logobj)
            LOG.debug("LogServer received new record: %s", record)
            logging.getLogger(record.name).handle(record)

    async def start(self):
        self._server = await asyncio.start_server(
            self._add_client,
            host=self._listen_host,
            port=self._listen_port,
        )
        await self._server.start_serving()

    async def stop(self):
        self._server.close()
        await self._server.wait_closed()
        for _, task in self._clients:
            task.cancel()



from typing import TYPE_CHECKING

from palaestrai.util.exception import (
    DeadChildrenRisingAsZombiesError,
    InvalidRequestError,
    RequestIsNoneError,
    SignalInterruptError,
    TasksNotFinishedError,
)

from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from palaestrai.experiment import RunGovernor


class RGSErrorHandlingTransceiving(RunGovernorState):
    """Represent the ERROR_HANDLING_TRANSCEIVING state of the run
    governor.

    Expected errors:
    * InvalidRequestError
    * TasksNotFinishedError
    * SignalInterruptError
    * DeadChildrenRisingAsZombiesError
    * RequestIsNoneError

     Possible next states are
    * :class:`.RGSTransceiving` in the InvalidRequestError,
        TasksNotFinishedError, RequestIsNoneError case.
    * :class:`.RGSStoppingSimulation` in the SignalInterruptError case.
    * :class:`.RGSHandlingDeadChildren` in the
        DeadChildrenRisingAsZombiesError case.

    """

    def __init__(self, rgc: "RunGovernor"):
        super().__init__(rgc, "ERROR_HANDLING_TRANSCEIVING")

    async def run(self):
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) encountered errors during "
            "transceiving. Handling these errors now.",
            id(self.rgc),
            self.rgc.uid,
        )
        for (error, state) in self.rgc.errors:
            LOG.debug(
                "RunGovernor(id=0x%x, uid=%s) encountered %s during "
                "state %s",
                id(self.rgc),
                self.rgc.uid,
                error.__class__,
                state,
            )

        assert len(self.rgc.errors) == 1

    def next_state(self):
        from . import (
            RGSHandlingDeadChildren,
            RGSStoppingSimulation,
            RGSTransceiving,
        )

        dispatcher = {
            InvalidRequestError: RGSTransceiving,
            TasksNotFinishedError: RGSTransceiving,
            SignalInterruptError: RGSStoppingSimulation,
            DeadChildrenRisingAsZombiesError: RGSHandlingDeadChildren,
            RequestIsNoneError: RGSTransceiving,
        }

        self.rgc.state = dispatcher[self.rgc.errors[0][0].__class__](self.rgc)
        self.rgc.errors = list()



from dataclasses import dataclass


@dataclass
class EnvironmentSetupRequest:
    """Instructs an :class:`EnvironmentConductor` to initialize
    its :class:`Environment`

    * Sender: :class:`SimulationController`
    * Receiver: :class:`EnvironmentConductor`

    :param environment_id: ID of the :class:`Environment` to initialize
    :param experiment_run_id: ID of the experiment run for which the
        environment is set up
    :param sender_simulation_controller_id: The sending
        :class:`SimulationController`
    :param receiver_environment_conductor_id: Target
        :class:`EnvironmentConductor`
    """

    environment_id: str
    experiment_run_id: str
    sender_simulation_controller_id: str
    receiver_environment_conductor_id: str

    @property
    def environment_conductor_id(self):
        return self.receiver_envionrment_conductor_id

    @property
    def sender_id(self):
        return self.sender_simulation_controller_id

    @property
    def receiver(self):
        return self.receiver_envionrment_conductor_id



import unittest
from copy import deepcopy
from pathlib import Path

from palaestrai.core.runtime_config import RuntimeConfig, _RuntimeConfig


class RuntimeConfigTestCase(unittest.TestCase):
    def test_singleton(self):
        self.assertEqual(id(RuntimeConfig()), id(RuntimeConfig()))

    def test_implicit_load(self):
        c = RuntimeConfig()
        self.assertTrue(c._loaded_configuration)

    def test_load_from_file(self):
        c = RuntimeConfig()
        self.assertTrue(c._config_file_path)
        config_file_path = (
            Path(__file__).parent
            / ".."
            / ".."
            / ".."
            / "fixtures"
            / "palaestrai-runtime-debug.conf.yaml"
        )
        c.load(config_file_path)
        self.assertTrue(c._config_file_path)
        self.assertEqual(c._config_file_path, config_file_path)

    def test_load_from_stream(self):
        from io import StringIO

        custom_store_uri = "psql://foo:bar@baz.example.com/meeple"
        stream = StringIO("---\nstore_uri: %s" % custom_store_uri)
        RuntimeConfig().reset()
        RuntimeConfig().load(stream)
        self.assertEqual(RuntimeConfig().store_uri, custom_store_uri)

    def test_modifiable_config_and_immutable_loaded_config(self):
        c = _RuntimeConfig()
        c._loaded_configuration = deepcopy(_RuntimeConfig.DEFAULT_CONFIG)
        self.assertIsNotNone(c.store_uri)
        self.assertEqual(c.store_uri, "sqlite:///palaestrai.db")
        self.assertEqual(
            c.store_uri, _RuntimeConfig.DEFAULT_CONFIG["store_uri"]
        )
        c._loaded_configuration["store_uri"] = None
        self.assertIsNotNone(c.store_uri)
        self.assertEqual(
            c.store_uri, _RuntimeConfig.DEFAULT_CONFIG["store_uri"]
        )

    def test_implicit_default_config(self):
        c = _RuntimeConfig()
        self.assertEqual(
            c._get("executor_bus_port"),
            _RuntimeConfig.DEFAULT_CONFIG["executor_bus_port"],
        )

    def test_load_from_dict(self):
        c = _RuntimeConfig()
        c.load({"store_uri": "Foo"})
        self.assertEqual(c.store_uri, "Foo")
        self.assertEqual(
            c.executor_bus_port,
            _RuntimeConfig.DEFAULT_CONFIG["executor_bus_port"],
        )

    def test_reset_with_dict(self):
        c = _RuntimeConfig()
        c.load({"store_uri": "Foo"})
        self.assertEqual(c.store_uri, "Foo")

        c.reset()
        self.assertIsNone(c._config_file_path)
        self.assertDictEqual(c._loaded_configuration, {})
        self.assertEqual(
            c.store_uri, _RuntimeConfig.DEFAULT_CONFIG["store_uri"]
        )

    def test_reset(self):
        c = _RuntimeConfig()
        self.assertEqual(
            c.store_uri, _RuntimeConfig.DEFAULT_CONFIG["store_uri"]
        )
        c._store_uri = None
        self.assertIsNone(c.store_uri)
        c.reset()
        self.assertEqual(
            c.store_uri, _RuntimeConfig.DEFAULT_CONFIG["store_uri"]
        )


if __name__ == "__main__":
    unittest.main()



from .termination_condition import TerminationCondition


class VanillaRunGovernorTerminationCondition(TerminationCondition):
    def check_termination(self, message, component):
        if component:
            return False
        return True



import unittest
from unittest.mock import AsyncMock, MagicMock, PropertyMock, patch

from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import (
    RGSFinalizing,
    RGSStoppingRun,
)


class TestStoppingRun(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)
        self.rgs = RGSStoppingRun(self.rungov)
        self.rungov.state = self.rgs

    async def test_run(self):
        self.rgs._stop_processes = AsyncMock()
        # self.rungov.major_domo_worker.transceive = AsyncMock()

        await self.rgs.run()

        self.rgs._stop_processes.assert_called_once()

    def test_next_state(self):
        self.rungov.experiment_run = MagicMock()
        self.rungov.experiment_run.has_next_phase = MagicMock(
            return_value=False
        )
        self.rgs.next_state()
        self.assertIsInstance(self.rungov.state, RGSFinalizing)

    @patch(f"{RGSStoppingRun.__module__}.asyncio.wait")
    @patch(f"{RGSStoppingRun.__module__}.asyncio.create_task")
    async def test_stop_processes(self, mock_wait, mock_create):
        self.rgs._join_process = MagicMock()
        self.rungov.sim_controllers["sim-1"] = 0
        await self.rgs._stop_processes()

        mock_wait.assert_called_once()
        mock_create.assert_called_once()
        self.rgs._join_process.assert_called_once()

    @patch(f"{RGSStoppingRun.__module__}.asyncio")
    async def test_join_process(self, mock_asyncio):
        proc = MagicMock()
        proc.name.return_value = "proc-1"
        proc.is_alive = MagicMock(return_value=True)
        type(proc).exitcode = PropertyMock(side_effect=[None, 1, 1, 1])
        mock_asyncio.sleep = AsyncMock()

        await self.rgs._join_process(proc)

        mock_asyncio.sleep.assert_called_once()
        self.assertEqual(2, proc.is_alive.call_count)


if __name__ == "__main__":
    unittest.main()
